from __future__ import annotations

from enum import Enum, IntEnum, IntFlag
from typing import Literal, NotRequired, Union

from typing_extensions import TypedDict

class SignatureHelp(TypedDict):
    """Signature help represents the signature of something
    callable. There can be multiple signature but only one
    active and only one active parameter.
    """

    signatures: list["SignatureInformation"]
    """ One or more signatures. """
    activeSignature: NotRequired[Uint]
    """ The active signature. If omitted or the value lies outside the
    range of `signatures` the value defaults to zero or is ignored if
    the `SignatureHelp` has no signatures.

    Whenever possible implementers should make an active decision about
    the active signature and shouldn't rely on a default value.

    In future version of the protocol this property might become
    mandatory to better express this. """
    activeParameter: NotRequired[Uint]
    """ The active parameter of the active signature. If omitted or the value
    lies outside the range of `signatures[activeSignature].parameters`
    defaults to 0 if the active signature has parameters. If
    the active signature has no parameters it is ignored.
    In future version of the protocol this property might become
    mandatory to better express the active parameter if the
    active signature does have any. """


class SignatureHelpRegistrationOptions(TypedDict):
    """Registration options for a {@link SignatureHelpRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    triggerCharacters: NotRequired[list[str]]
    """ List of characters that trigger signature help automatically. """
    retriggerCharacters: NotRequired[list[str]]
    """ List of characters that re-trigger signature help.

    These trigger characters are only active when signature help is already showing. All trigger characters
    are also counted as re-trigger characters.

    @since 3.15.0 """


class DefinitionParams(TypedDict):
    """Parameters for a {@link DefinitionRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class DefinitionRegistrationOptions(TypedDict):
    """Registration options for a {@link DefinitionRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class ReferenceParams(TypedDict):
    """Parameters for a {@link ReferencesRequest}."""

    context: "ReferenceContext"
    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class ReferenceRegistrationOptions(TypedDict):
    """Registration options for a {@link ReferencesRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class DocumentHighlightParams(TypedDict):
    """Parameters for a {@link DocumentHighlightRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class DocumentHighlight(TypedDict):
    """A document highlight is a range inside a text document which deserves
    special attention. Usually a document highlight is visualized by changing
    the background color of its range.
    """

    range: "Range"
    """ The range this highlight applies to. """
    kind: NotRequired["DocumentHighlightKind"]
    """ The highlight kind, default is {@link DocumentHighlightKind.Text text}. """


class DocumentHighlightRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentHighlightRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class DocumentSymbolParams(TypedDict):
    """Parameters for a {@link DocumentSymbolRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class SymbolInformation(TypedDict):
    """Represents information about programming constructs like variables, classes,
    interfaces etc.
    """

    deprecated: NotRequired[bool]
    """ Indicates if this symbol is deprecated.

    @deprecated Use tags instead """
    location: "Location"
    """ The location of this symbol. The location's range is used by a tool
    to reveal the location in the editor. If the symbol is selected in the
    tool the range's start information is used to position the cursor. So
    the range usually spans more than the actual symbol's name and does
    normally include things like visibility modifiers.

    The range doesn't have to denote a node range in the sense of an abstract
    syntax tree. It can therefore not be used to re-construct a hierarchy of
    the symbols. """
    name: str
    """ The name of this symbol. """
    kind: "SymbolKind"
    """ The kind of this symbol. """
    tags: NotRequired[list["SymbolTag"]]
    """ Tags for this symbol.

    @since 3.16.0 """
    containerName: NotRequired[str]
    """ The name of the symbol containing this symbol. This information is for
    user interface purposes (e.g. to render a qualifier in the user interface
    if necessary). It can't be used to re-infer a hierarchy for the document
    symbols. """


class DocumentSymbol(TypedDict):
    """Represents programming constructs like variables, classes, interfaces etc.
    that appear in a document. Document symbols can be hierarchical and they
    have two ranges: one that encloses its definition and one that points to
    its most interesting range, e.g. the range of an identifier.
    """

    name: str
    """ The name of this symbol. Will be displayed in the user interface and therefore must not be
    an empty string or a string only consisting of white spaces. """
    detail: NotRequired[str]
    """ More detail for this symbol, e.g the signature of a function. """
    kind: "SymbolKind"
    """ The kind of this symbol. """
    tags: NotRequired[list["SymbolTag"]]
    """ Tags for this document symbol.

    @since 3.16.0 """
    deprecated: NotRequired[bool]
    """ Indicates if this symbol is deprecated.

    @deprecated Use tags instead """
    range: "Range"
    """ The range enclosing this symbol not including leading/trailing whitespace but everything else
    like comments. This information is typically used to determine if the clients cursor is
    inside the symbol to reveal in the symbol in the UI. """
    selectionRange: "Range"
    """ The range that should be selected and revealed when this symbol is being picked, e.g the name of a function.
    Must be contained by the `range`. """

    # NOTE: 원문 스펙 대비 children 필드 누락 가능성에 대한 메모


class DocumentSymbolRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentSymbolRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    label: NotRequired[str]
    """ A human-readable string that is shown when multiple outlines trees
    are shown for the same document.

    @since 3.16.0 """


class CodeActionParams(TypedDict):
    """The parameters of a {@link CodeActionRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document in which the command was invoked. """
    range: "Range"
    """ The range for which the command was invoked. """
    context: "CodeActionContext"
    """ Context carrying additional information. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class Command(TypedDict):
    """Represents a reference to a command. Provides a title which
    will be used to represent a command in the UI and, optionally,
    an array of arguments which will be passed to the command handler
    function when invoked.
    """

    title: str
    """ Title of the command, like `save`. """
    command: str
    """ The identifier of the actual command handler. """
    arguments: NotRequired[list["LSPAny"]]
    """ Arguments that the command handler should be
    invoked with. """


class CodeAction(TypedDict):
    """A code action represents a change that can be performed in code, e.g. to fix a problem or
    to refactor code.

    A CodeAction must set either `edit` and/or a `command`. If both are supplied, the `edit` is applied first, then the `command` is executed.
    """

    title: str
    """ A short, human-readable, title for this code action. """
    kind: NotRequired["CodeActionKind"]
    """ The kind of the code action.

    Used to filter code actions. """
    diagnostics: NotRequired[list["Diagnostic"]]
    """ The diagnostics that this code action resolves. """
    isPreferred: NotRequired[bool]
    """ Marks this as a preferred action. Preferred actions are used by the `auto fix` command and can be targeted
    by keybindings.

    A quick fix should be marked preferred if it properly addresses the underlying error.
    A refactoring should be marked preferred if it is the most reasonable choice of actions to take.

    @since 3.15.0 """
    disabled: NotRequired["__CodeAction_disabled_Type_1"]
    """ Marks that the code action cannot currently be applied.

    Clients should follow the following guidelines regarding disabled code actions:

      - Disabled code actions are not shown in automatic [lightbulbs](https://code.visualstudio.com/docs/editor/editingevolved#_code-action)
        code action menus.

      - Disabled actions are shown as faded out in the code action menu when the user requests a more specific type
        of code action, such as refactorings.

      - If the user has a [keybinding](https://code.visualstudio.com/docs/editor/refactoring#_keybindings-for-code-actions)
        that auto applies a code action and only disabled code actions are returned, the client should show the user an
        error message with `reason` in the editor.

    @since 3.16.0 """
    edit: NotRequired["WorkspaceEdit"]
    """ The workspace edit this code action performs. """
    command: NotRequired["Command"]
    """ A command this code action executes. If a code action
    provides an edit and a command, first the edit is
    executed and then the command. """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on a code action between
    a `textDocument/codeAction` and a `codeAction/resolve` request.

    @since 3.16.0 """


class CodeActionRegistrationOptions(TypedDict):
    """Registration options for a {@link CodeActionRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    codeActionKinds: NotRequired[list["CodeActionKind"]]
    """ CodeActionKinds that this server may return.

    The list of kinds may be generic, such as `CodeActionKind.Refactor`, or the server
    may list out every specific kind they provide. """
    resolveProvider: NotRequired[bool]
    """ The server provides support to resolve additional
    information for a code action.

    @since 3.16.0 """


class WorkspaceSymbolParams(TypedDict):
    """The parameters of a {@link WorkspaceSymbolRequest}."""

    query: str
    """ A query string to filter symbols by. Clients may send an empty
    string here to request all symbols. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class WorkspaceSymbol(TypedDict):
    """A special workspace symbol that supports locations without a range.

    See also SymbolInformation.

    @since 3.17.0
    """

    location: Union["Location", "__WorkspaceSymbol_location_Type_1"]
    """ The location of the symbol. Whether a server is allowed to
    return a location without a range depends on the client
    capability `workspace.symbol.resolveSupport`.

    See SymbolInformation#location for more details. """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on a workspace symbol between a
    workspace symbol request and a workspace symbol resolve request. """
    name: str
    """ The name of this symbol. """
    kind: "SymbolKind"
    """ The kind of this symbol. """
    tags: NotRequired[list["SymbolTag"]]
    """ Tags for this symbol.

    @since 3.16.0 """
    containerName: NotRequired[str]
    """ The name of the symbol containing this symbol. This information is for
    user interface purposes (e.g. to render a qualifier in the user interface
    if necessary). It can't be used to re-infer a hierarchy for the document
    symbols. """


class WorkspaceSymbolRegistrationOptions(TypedDict):
    """Registration options for a {@link WorkspaceSymbolRequest}."""

    resolveProvider: NotRequired[bool]
    """ The server provides support to resolve additional
    information for a workspace symbol.

    @since 3.17.0 """


class CodeLensParams(TypedDict):
    """The parameters of a {@link CodeLensRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to request code lens for. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class CodeLens(TypedDict):
    """A code lens represents a {@link Command command} that should be shown along with
    source text, like the number of references, a way to run tests, etc.

    A code lens is _unresolved_ when no command is associated to it. For performance
    reasons the creation of a code lens and resolving should be done in two stages.
    """

    range: "Range"
    """ The range in which this code lens is valid. Should only span a single line. """
    command: NotRequired["Command"]
    """ The command this code lens represents. """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on a code lens item between
    a {@link CodeLensRequest} and a [CodeLensResolveRequest]
    (#CodeLensResolveRequest) """


class CodeLensRegistrationOptions(TypedDict):
    """Registration options for a {@link CodeLensRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    resolveProvider: NotRequired[bool]
    """ Code lens has a resolve provider as well. """


class DocumentLinkParams(TypedDict):
    """The parameters of a {@link DocumentLinkRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to provide document links for. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class DocumentLink(TypedDict):
    """A document link is a range in a text document that links to an internal or external resource, like another
    text document or a web site.
    """

    range: "Range"
    """ The range this link applies to. """
    target: NotRequired[str]
    """ The uri this link points to. If missing a resolve request is sent later. """
    tooltip: NotRequired[str]
    """ The tooltip text when you hover over this link.

    If a tooltip is provided, is will be displayed in a string that includes instructions on how to
    trigger the link, such as `{0} (ctrl + click)`. The specific instructions vary depending on OS,
    user settings, and localization.

    @since 3.15.0 """
    data: NotRequired["LSPAny"]
    """ A data entry field that is preserved on a document link between a
    DocumentLinkRequest and a DocumentLinkResolveRequest. """


class DocumentLinkRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentLinkRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    resolveProvider: NotRequired[bool]
    """ Document links have a resolve provider as well. """


class DocumentFormattingParams(TypedDict):
    """The parameters of a {@link DocumentFormattingRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to format. """
    options: "FormattingOptions"
    """ The format options. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class DocumentFormattingRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentFormattingRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class DocumentRangeFormattingParams(TypedDict):
    """The parameters of a {@link DocumentRangeFormattingRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to format. """
    range: "Range"
    """ The range to format """
    options: "FormattingOptions"
    """ The format options """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class DocumentRangeFormattingRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentRangeFormattingRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """


class DocumentOnTypeFormattingParams(TypedDict):
    """The parameters of a {@link DocumentOnTypeFormattingRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to format. """
    position: "Position"
    """ The position around which the on type formatting should happen.
    This is not necessarily the exact position where the character denoted
    by the property `ch` got typed. """
    ch: str
    """ The character that has been typed that triggered the formatting
    on type request. That is not necessarily the last character that
    got inserted into the document since the client could auto insert
    characters as well (e.g. like automatic brace completion). """
    options: "FormattingOptions"
    """ The formatting options. """


class DocumentOnTypeFormattingRegistrationOptions(TypedDict):
    """Registration options for a {@link DocumentOnTypeFormattingRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    firstTriggerCharacter: str
    """ A character on which formatting should be triggered, like `{`. """
    moreTriggerCharacter: NotRequired[list[str]]
    """ More trigger characters. """


class RenameParams(TypedDict):
    """The parameters of a {@link RenameRequest}."""

    textDocument: "TextDocumentIdentifier"
    """ The document to rename. """
    position: "Position"
    """ The position at which this request was sent. """
    newName: str
    """ The new name of the symbol. If the given name is not valid the
    request must return a {@link ResponseError} with an
    appropriate message set. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class RenameRegistrationOptions(TypedDict):
    """Registration options for a {@link RenameRequest}."""

    documentSelector: Union["DocumentSelector", None]
    """ A document selector to identify the scope of the registration. If set to null
    the document selector provided on the client side will be used. """
    prepareProvider: NotRequired[bool]
    """ Renames should be checked and tested before being executed.

    @since version 3.12.0 """


class PrepareRenameParams(TypedDict):
    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class ExecuteCommandParams(TypedDict):
    """The parameters of a {@link ExecuteCommandRequest}."""

    command: str
    """ The identifier of the actual command handler. """
    arguments: NotRequired[list["LSPAny"]]
    """ Arguments that the command should be invoked with. """
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class ExecuteCommandRegistrationOptions(TypedDict):
    """Registration options for a {@link ExecuteCommandRequest}."""

    commands: list[str]
    """ The commands to be executed on the server """


class ApplyWorkspaceEditParams(TypedDict):
    """The parameters passed via a apply workspace edit request."""

    label: NotRequired[str]
    """ An optional label of the workspace edit. This label is
    presented in the user interface for example on an undo
    stack to undo the workspace edit. """
    edit: "WorkspaceEdit"
    """ The edits to apply. """


class ApplyWorkspaceEditResult(TypedDict):
    """The result returned from the apply workspace edit request.

    @since 3.17 renamed from ApplyWorkspaceEditResponse
    """

    applied: bool
    """ Indicates whether the edit was applied or not. """
    failureReason: NotRequired[str]
    """ An optional textual description for why the edit was not applied.
    This may be used by the server for diagnostic logging or to provide
    a suitable error for a request that triggered the edit. """
    failedChange: NotRequired[Uint]
    """ Depending on the client's failure handling strategy `failedChange` might
    contain the index of the change that failed. This property is only available
    if the client signals a `failureHandlingStrategy` in its client capabilities. """


class WorkDoneProgressBegin(TypedDict):
    kind: Literal["begin"]
    title: str
    """ Mandatory title of the progress operation. Used to briefly inform about
    the kind of operation being performed.

    Examples: "Indexing" or "Linking dependencies". """
    cancellable: NotRequired[bool]
    """ Controls if a cancel button should show to allow the user to cancel the
    long running operation. Clients that don't support cancellation are allowed
    to ignore the setting. """
    message: NotRequired[str]
    """ Optional, more detailed associated progress message. Contains
    complementary information to the `title`.

    Examples: "3/25 files", "project/src/module2", "node_modules/some_dep".
    If unset, the previous progress message (if any) is still valid. """
    percentage: NotRequired[Uint]
    """ Optional progress percentage to display (value 100 is considered 100%).
    If not provided infinite progress is assumed and clients are allowed
    to ignore the `percentage` value in subsequent in report notifications.

    The value should be steadily rising. Clients are free to ignore values
    that are not following this rule. The value range is [0, 100]. """


class WorkDoneProgressReport(TypedDict):
    kind: Literal["report"]
    cancellable: NotRequired[bool]
    """ Controls enablement state of a cancel button.

    Clients that don't support cancellation or don't support controlling the button's
    enablement state are allowed to ignore the property. """
    message: NotRequired[str]
    """ Optional, more detailed associated progress message. Contains
    complementary information to the `title`.

    Examples: "3/25 files", "project/src/module2", "node_modules/some_dep".
    If unset, the previous progress message (if any) is still valid. """
    percentage: NotRequired[Uint]
    """ Optional progress percentage to display (value 100 is considered 100%).
    If not provided infinite progress is assumed and clients are allowed
    to ignore the `percentage` value in subsequent in report notifications.

    The value should be steadily rising. Clients are free to ignore values
    that are not following this rule. The value range is [0, 100] """


class WorkDoneProgressEnd(TypedDict):
    kind: Literal["end"]
    message: NotRequired[str]
    """ Optional, a final message indicating to for example indicate the outcome
    of the operation. """


class SetTraceParams(TypedDict):
    value: "TraceValues"


class LogTraceParams(TypedDict):
    message: str
    verbose: NotRequired[str]


class CancelParams(TypedDict):
    id: int | str
    """ The request id to cancel. """


class ProgressParams(TypedDict):
    token: "ProgressToken"
    """ The progress token provided by the client or server. """
    value: "LSPAny"
    """ The progress data. """


class TextDocumentPositionParams(TypedDict):
    """A parameter literal used in requests to pass a text document and a position inside that
    document.
    """

    textDocument: "TextDocumentIdentifier"
    """ The text document. """
    position: "Position"
    """ The position inside the text document. """


class WorkDoneProgressParams(TypedDict):
    workDoneToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report work done progress. """


class PartialResultParams(TypedDict):
    partialResultToken: NotRequired["ProgressToken"]
    """ An optional token that a server can use to report partial results (e.g. streaming) to
    the client. """


class LocationLink(TypedDict):
    """Represents the connection of two locations. Provides additional metadata over normal {@link Location locations},
    including an origin range.
    """

    originSelectionRange: NotRequired["Range"]
    """ Span of the origin of this link.

    Used as the underlined span for mouse interaction. Defaults to the word range at
    the definition position. """
    targetUri: "DocumentUri"
    """ The target resource identifier of this link. """
    targetRange: "Range"
    """ The full target range of this link. If the target for example is a symbol then target range is the
    range enclosing this symbol not including leading/trailing whitespace but everything else
    like comments. This information is typically used to highlight the range in the editor. """
    targetSelectionRange: "Range"
    """ The range that should be selected and revealed when this link is being followed, e.g the name of a function.
    Must be contained by the `targetRange`. See also `DocumentSymbol#range` """


class Range(TypedDict):
    """A range in a text document expressed as (zero-based) start and end positions.

    If you want to specify a range that contains a line including the line ending
    character(s) then use an end position denoting the start of the next line.
    For example:
    ```ts
    {
        start: { line: 5, character: 23 }
        end : { line 6, character : 0 }
    }
    ```
    """

    start: "Position"
    """ The range's start position. """
    end: "Position"
    """ The range's end position. """


class ImplementationOptions(TypedDict):
    workDoneProgress: NotRequired[bool]





"""MCP daemon forward framed 통신 유틸을 제공한다."""

from __future__ import annotations

import json
import http.client


JsonMap = dict[str, object]


class DaemonForwardError(RuntimeError):
    """데몬 포워딩 실패를 나타낸다."""


def forward_once(request: JsonMap, host: str, port: int, timeout_sec: float) -> JsonMap:
    """단일 MCP 요청을 daemon endpoint로 전달한다."""
    body = json.dumps(request).encode("utf-8")
    connection = http.client.HTTPConnection(host, port, timeout=timeout_sec)
    try:
        connection.request("POST", "/mcp", body=body, headers={"Content-Type": "application/json"})
        response = connection.getresponse()
        response_body = response.read()
    finally:
        connection.close()
    if response.status < 200 or response.status >= 300:
        raise DaemonForwardError(f"daemon http status={response.status}")
    try:
        decoded = json.loads(response_body.decode("utf-8"))
    except (UnicodeDecodeError, json.JSONDecodeError) as exc:
        raise DaemonForwardError("daemon response must be json object") from exc
    if not isinstance(decoded, dict):
        raise DaemonForwardError("daemon response must be json object")
    return decoded



"""벡터 임베딩 저장소를 구현한다."""

from __future__ import annotations

import json
from pathlib import Path

from sari.db.row_mapper import row_str
from sari.db.schema import connect


class VectorEmbeddingRepository:
    """파일/질의 벡터 임베딩 영속화를 담당한다."""

    def __init__(self, db_path: Path) -> None:
        """저장소에 사용할 DB 경로를 저장한다."""
        self._db_path = db_path

    def upsert_file_embedding(
        self,
        repo_root: str,
        relative_path: str,
        content_hash: str,
        model_id: str,
        dim: int,
        vector: list[float],
        updated_at: str,
    ) -> None:
        """파일 임베딩을 업서트한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                INSERT INTO file_embeddings(
                    repo_root, relative_path, content_hash, model_id, dim, vector_json, updated_at
                )
                VALUES(
                    :repo_root, :relative_path, :content_hash, :model_id, :dim, :vector_json, :updated_at
                )
                ON CONFLICT(repo_root, relative_path, content_hash, model_id) DO UPDATE SET
                    dim = excluded.dim,
                    vector_json = excluded.vector_json,
                    updated_at = excluded.updated_at
                """,
                {
                    "repo_root": repo_root,
                    "relative_path": relative_path,
                    "content_hash": content_hash,
                    "model_id": model_id,
                    "dim": dim,
                    "vector_json": json.dumps(vector, ensure_ascii=False),
                    "updated_at": updated_at,
                },
            )
            conn.commit()

    def get_file_embedding(
        self,
        repo_root: str,
        relative_path: str,
        content_hash: str,
        model_id: str,
    ) -> list[float] | None:
        """파일 임베딩을 조회한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                """
                SELECT vector_json
                FROM file_embeddings
                WHERE repo_root = :repo_root
                  AND relative_path = :relative_path
                  AND content_hash = :content_hash
                  AND model_id = :model_id
                """,
                {
                    "repo_root": repo_root,
                    "relative_path": relative_path,
                    "content_hash": content_hash,
                    "model_id": model_id,
                },
            ).fetchone()
        if row is None:
            return None
        loaded = json.loads(row_str(row, "vector_json"))
        if not isinstance(loaded, list):
            return None
        output: list[float] = []
        for value in loaded:
            if isinstance(value, (int, float)):
                output.append(float(value))
        return output

    def upsert_query_embedding(self, query_hash: str, model_id: str, dim: int, vector: list[float], updated_at: str) -> None:
        """질의 임베딩을 업서트한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                INSERT INTO query_embeddings(query_hash, model_id, dim, vector_json, updated_at)
                VALUES(:query_hash, :model_id, :dim, :vector_json, :updated_at)
                ON CONFLICT(query_hash, model_id) DO UPDATE SET
                    dim = excluded.dim,
                    vector_json = excluded.vector_json,
                    updated_at = excluded.updated_at
                """,
                {
                    "query_hash": query_hash,
                    "model_id": model_id,
                    "dim": dim,
                    "vector_json": json.dumps(vector, ensure_ascii=False),
                    "updated_at": updated_at,
                },
            )
            conn.commit()

    def get_query_embedding(self, query_hash: str, model_id: str) -> list[float] | None:
        """질의 임베딩을 조회한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                """
                SELECT vector_json
                FROM query_embeddings
                WHERE query_hash = :query_hash
                  AND model_id = :model_id
                """,
                {"query_hash": query_hash, "model_id": model_id},
            ).fetchone()
        if row is None:
            return None
        loaded = json.loads(row_str(row, "vector_json"))
        if not isinstance(loaded, list):
            return None
        output: list[float] = []
        for value in loaded:
            if isinstance(value, (int, float)):
                output.append(float(value))
        return output



"""세션 단위 stabilization 메트릭 상태를 관리한다."""

from __future__ import annotations

import os
import threading
from dataclasses import dataclass
from typing import Mapping

from sari.mcp.stabilization.analytics_queue import enqueue_analytics
from sari.mcp.stabilization.session_keys import resolve_session_key, strict_session_id_enabled


@dataclass
class _SessionMetrics:
    """세션 메트릭 누적 상태를 표현한다."""

    reads_count: int = 0
    reads_lines_total: int = 0
    reads_chars_total: int = 0
    search_count: int = 0
    read_span_sum: int = 0
    max_read_span: int = 0
    preview_degraded_count: int = 0
    reads_after_search_count: int = 0
    last_search_query: str = ""
    last_search_top_paths: tuple[str, ...] = ()
    last_search_candidates: dict[str, str] | None = None
    last_bundle_id: str = ""
    last_seen_seq: int = 0


_LOCK = threading.RLock()
_SESSION_METRICS: dict[str, _SessionMetrics] = {}
_SEQUENCE = 0
_MAX_SESSION_METRICS = max(64, int(os.environ.get("SARI_SESSION_METRICS_MAX", "4096") or "4096"))


def _next_sequence() -> int:
    """증분 시퀀스를 반환한다."""
    global _SEQUENCE
    _SEQUENCE += 1
    return _SEQUENCE


def _session_key(args: Mapping[str, object] | object, roots: list[str]) -> str:
    """요청별 세션 키를 계산한다."""
    return resolve_session_key(args, roots)


def _evict_metrics_if_needed() -> None:
    """메트릭 캐시 상한을 넘으면 오래된 항목부터 제거한다."""
    if len(_SESSION_METRICS) < _MAX_SESSION_METRICS:
        return
    overflow = len(_SESSION_METRICS) - _MAX_SESSION_METRICS + 1
    victims = sorted(_SESSION_METRICS.items(), key=lambda item: int(item[1].last_seen_seq or 0))[:overflow]
    for key, _state in victims:
        _SESSION_METRICS.pop(key, None)


def _state_of(session_key: str) -> _SessionMetrics:
    """세션 키에 대응하는 메트릭 상태를 반환한다."""
    state = _SESSION_METRICS.get(session_key)
    if state is None:
        _evict_metrics_if_needed()
        state = _SessionMetrics()
        _SESSION_METRICS[session_key] = state
    return state


def _snapshot(state: _SessionMetrics) -> dict[str, float | int]:
    """메트릭 상태를 외부 응답용 스냅샷으로 변환한다."""
    ratio = (state.reads_after_search_count / state.reads_count) if state.reads_count > 0 else 0.0
    avg_span = (state.read_span_sum / state.reads_count) if state.reads_count > 0 else 0.0
    return {
        "reads_count": state.reads_count,
        "reads_lines_total": state.reads_lines_total,
        "reads_chars_total": state.reads_chars_total,
        "search_count": state.search_count,
        "read_after_search_ratio": round(ratio, 6),
        "avg_read_span": round(avg_span, 6),
        "max_read_span": state.max_read_span,
        "preview_degraded_count": state.preview_degraded_count,
    }


def _enqueue_analytics_snapshot(event_type: str, session_key: str, state: _SessionMetrics) -> None:
    """분석 큐에 스냅샷 이벤트를 넣는다."""
    enqueue_analytics(
        {
            "event_type": event_type,
            "session_key": session_key,
            "snapshot": _snapshot(state),
            "seq": state.last_seen_seq,
        }
    )


def record_search_metrics(
    args: Mapping[str, object] | object,
    roots: list[str],
    preview_degraded: bool,
    query: str = "",
    top_paths: list[str] | None = None,
    candidates: Mapping[str, str] | None = None,
    bundle_id: str = "",
) -> dict[str, float | int]:
    """search 메트릭을 누적하고 최신 스냅샷을 반환한다."""
    key = _session_key(args, roots)
    with _LOCK:
        state = _state_of(key)
        state.search_count += 1
        state.last_seen_seq = _next_sequence()
        state.last_search_query = str(query or "").strip()
        if top_paths is not None:
            state.last_search_top_paths = tuple(str(path) for path in top_paths if str(path).strip() != "")
        if candidates is not None:
            state.last_search_candidates = {str(k): str(v) for k, v in candidates.items()}
        if bundle_id != "":
            state.last_bundle_id = str(bundle_id)
        if preview_degraded:
            state.preview_degraded_count += 1
        _enqueue_analytics_snapshot("search", key, state)
        return _snapshot(state)


def record_read_metrics(
    args: Mapping[str, object] | object,
    roots: list[str],
    read_lines: int,
    read_chars: int,
    read_span: int,
) -> dict[str, float | int]:
    """read 메트릭을 누적하고 최신 스냅샷을 반환한다."""
    key = _session_key(args, roots)
    with _LOCK:
        state = _state_of(key)
        state.reads_count += 1
        state.last_seen_seq = _next_sequence()
        safe_lines = max(0, int(read_lines))
        safe_chars = max(0, int(read_chars))
        safe_span = max(0, int(read_span))
        state.reads_lines_total += safe_lines
        state.reads_chars_total += safe_chars
        state.read_span_sum += safe_span
        state.max_read_span = max(state.max_read_span, safe_span)
        if state.search_count > 0:
            state.reads_after_search_count += 1
        _enqueue_analytics_snapshot("read", key, state)
        return _snapshot(state)


def get_metrics_snapshot(args: Mapping[str, object] | object, roots: list[str]) -> dict[str, float | int]:
    """세션 메트릭 스냅샷을 반환한다."""
    key = _session_key(args, roots)
    with _LOCK:
        return _snapshot(_state_of(key))


def get_session_key(args: Mapping[str, object] | object, roots: list[str]) -> str:
    """요청 인자에서 계산된 세션 키를 반환한다."""
    return _session_key(args, roots)


def get_search_context(args: Mapping[str, object] | object, roots: list[str]) -> dict[str, object]:
    """최근 search 컨텍스트를 반환한다."""
    key = _session_key(args, roots)
    with _LOCK:
        state = _state_of(key)
        return {
            "last_search_query": state.last_search_query,
            "last_search_top_paths": list(state.last_search_top_paths),
            "last_search_candidates": dict(state.last_search_candidates or {}),
            "last_bundle_id": state.last_bundle_id,
            "search_count": state.search_count,
        }


def requires_strict_session_id(args: Mapping[str, object] | object) -> bool:
    """strict session 정책에서 session_id 누락 여부를 반환한다."""
    if not strict_session_id_enabled():
        return False
    args_map = args if isinstance(args, Mapping) else {}
    session_id = str(args_map.get("session_id") or "").strip()
    return session_id == ""


def reset_session_metrics_for_tests() -> None:
    """테스트를 위해 세션 메트릭 상태를 초기화한다."""
    global _SEQUENCE
    with _LOCK:
        _SESSION_METRICS.clear()
        _SEQUENCE = 0




from __future__ import annotations

from enum import Enum, IntEnum, IntFlag
from typing import Literal, NotRequired, Union

from typing_extensions import TypedDict

class __WorkspaceSymbolClientCapabilities_tagSupport_Type_1(TypedDict):
    valueSet: list["SymbolTag"]
    """ The tags supported by the client. """


class __WorkspaceSymbol_location_Type_1(TypedDict):
    uri: "DocumentUri"


class ___InitializeParams_clientInfo_Type_1(TypedDict):
    name: str
    """ The name of the client as defined by the client. """
    version: NotRequired[str]
    """ The client's version as defined by the client. """



"""파일 수집 런타임 포트 인터페이스를 정의한다."""

from __future__ import annotations

from typing import Protocol

from sari.core.models import CollectionScanResultDTO, FileReadResultDTO, PipelineMetricsDTO


class CollectionScanPort(Protocol):
    """L1 스캔/파일 조회 계층 포트를 정의한다."""

    def scan_once(self, repo_root: str) -> CollectionScanResultDTO:
        """저장소 전체를 1회 스캔한다."""
        ...

    def index_file(self, repo_root: str, relative_path: str) -> CollectionScanResultDTO:
        """단일 파일을 증분 인덱싱한다."""
        ...

    def list_files(self, repo_root: str, limit: int, prefix: str | None) -> list[dict[str, object]]:
        """저장소 내 인덱싱 파일 목록을 조회한다."""
        ...

    def read_file(self, repo_root: str, relative_path: str, offset: int, limit: int | None) -> FileReadResultDTO:
        """인덱싱 파일 본문을 읽는다."""
        ...


class CollectionPipelinePort(Protocol):
    """L2/L3 파이프라인 실행 포트를 정의한다."""

    def process_enrich_jobs(self, limit: int) -> int:
        """L2/L3 통합 보강 작업을 수행한다."""
        ...

    def process_enrich_jobs_l2(self, limit: int) -> int:
        """L2 보강 작업을 수행한다."""
        ...

    def process_enrich_jobs_l3(self, limit: int) -> int:
        """L3 보강 작업을 수행한다."""
        ...


class CollectionLifecyclePort(Protocol):
    """백그라운드 수집 런타임 생명주기 포트를 정의한다."""

    def start_background(self) -> None:
        """백그라운드 루프를 시작한다."""
        ...

    def stop_background(self) -> None:
        """백그라운드 루프를 중지한다."""
        ...


class CollectionObservabilityPort(Protocol):
    """파이프라인 관측/오류 조회 포트를 정의한다."""

    def get_pipeline_metrics(self) -> PipelineMetricsDTO:
        """파이프라인 메트릭을 조회한다."""
        ...

    def list_error_events(
        self,
        limit: int,
        offset: int = 0,
        repo_root: str | None = None,
        error_code: str | None = None,
    ) -> list[dict[str, object]]:
        """오류 이벤트 목록을 조회한다."""
        ...

    def get_error_event(self, event_id: str) -> dict[str, object] | None:
        """오류 이벤트 단건을 조회한다."""
        ...


class CollectionRuntimePort(
    CollectionScanPort,
    CollectionPipelinePort,
    CollectionLifecyclePort,
    CollectionObservabilityPort,
    Protocol,
):
    """수집 런타임 전체 기능을 제공하는 통합 포트다."""




"""
Provides Bash specific instantiation of the LanguageServer class using bash-language-server.
Contains various configurations and settings specific to Bash scripting.
"""

import logging
import os
import pathlib
import shutil
import threading

from solidlsp.language_servers.common import RuntimeDependency, RuntimeDependencyCollection
from solidlsp.ls import DocumentSymbols, LSPFileBuffer, SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class BashLanguageServer(SolidLanguageServer):
    """
    Provides Bash specific instantiation of the LanguageServer class using bash-language-server.
    Contains various configurations and settings specific to Bash scripting.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a BashLanguageServer instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        bash_lsp_executable_path = self._setup_runtime_dependencies(config, solidlsp_settings)
        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=bash_lsp_executable_path, cwd=repository_root_path),
            "bash",
            solidlsp_settings,
        )
        self.server_ready = threading.Event()
        self.initialize_searcher_command_available = threading.Event()

    @classmethod
    def _setup_runtime_dependencies(cls, config: LanguageServerConfig, solidlsp_settings: SolidLSPSettings) -> str:
        """
        Setup runtime dependencies for Bash Language Server and return the command to start the server.
        """
        # Verify both node and npm are installed
        is_node_installed = shutil.which("node") is not None
        assert is_node_installed, "node is not installed or isn't in PATH. Please install NodeJS and try again."
        is_npm_installed = shutil.which("npm") is not None
        assert is_npm_installed, "npm is not installed or isn't in PATH. Please install npm and try again."

        deps = RuntimeDependencyCollection(
            [
                RuntimeDependency(
                    id="bash-language-server",
                    description="bash-language-server package",
                    command="npm install --prefix ./ bash-language-server@5.6.0",
                    platform_id="any",
                ),
            ]
        )

        # Install bash-language-server if not already installed
        bash_ls_dir = os.path.join(cls.ls_resources_dir(solidlsp_settings), "bash-lsp")
        bash_executable_path = os.path.join(bash_ls_dir, "node_modules", ".bin", "bash-language-server")

        # Handle Windows executable extension
        if os.name == "nt":
            bash_executable_path += ".cmd"

        if not os.path.exists(bash_executable_path):
            log.info(f"Bash Language Server executable not found at {bash_executable_path}. Installing...")
            deps.install(bash_ls_dir)
            log.info("Bash language server dependencies installed successfully")

        if not os.path.exists(bash_executable_path):
            raise FileNotFoundError(
                f"bash-language-server executable not found at {bash_executable_path}, something went wrong with the installation."
            )
        return f"{bash_executable_path} start"

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Bash Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params = {
            "locale": "en",
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True, "dynamicRegistration": True},
                    "completion": {"dynamicRegistration": True, "completionItem": {"snippetSupport": True}},
                    "definition": {"dynamicRegistration": True},
                    "references": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                    },
                    "hover": {"dynamicRegistration": True, "contentFormat": ["markdown", "plaintext"]},
                    "signatureHelp": {"dynamicRegistration": True},
                    "codeAction": {"dynamicRegistration": True},
                },
                "workspace": {
                    "workspaceFolders": True,
                    "didChangeConfiguration": {"dynamicRegistration": True},
                    "symbol": {"dynamicRegistration": True},
                },
            },
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return initialize_params  # type: ignore

    def _start_server(self) -> None:
        """
        Starts the Bash Language Server, waits for the server to be ready and yields the LanguageServer instance.
        """

        def register_capability_handler(params: dict) -> None:
            assert "registrations" in params
            for registration in params["registrations"]:
                if registration["method"] == "workspace/executeCommand":
                    self.initialize_searcher_command_available.set()
            return

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")
            # Check for bash-language-server ready signals
            message_text = msg.get("message", "")
            if "Analyzing" in message_text or "analysis complete" in message_text.lower():
                log.info("Bash language server analysis signals detected")
                self.server_ready.set()

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting Bash server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)
        log.debug(f"Received initialize response from bash server: {init_response}")

        # Enhanced capability checks for bash-language-server 5.6.0
        assert init_response["capabilities"]["textDocumentSync"] in [1, 2]  # Full or Incremental
        assert "completionProvider" in init_response["capabilities"]

        # Verify document symbol support is available
        if "documentSymbolProvider" in init_response["capabilities"]:
            log.info("Bash server supports document symbols")
        else:
            log.warning("Warning: Bash server does not report document symbol support")

        self.server.notify.initialized({})

        # Wait for server readiness with timeout
        log.info("Waiting for Bash language server to be ready...")
        if not self.server_ready.wait(timeout=3.0):
            # Fallback: assume server is ready after timeout
            # This is common. bash-language-server doesn't always send explicit ready signals. Log as info
            log.info("Timeout waiting for bash server ready signal, proceeding anyway")
            self.server_ready.set()
        else:
            log.info("Bash server initialization complete")

    def request_document_symbols(self, relative_file_path: str, file_buffer: LSPFileBuffer | None = None) -> DocumentSymbols:
        # Uses the standard LSP documentSymbol request which provides reliable function detection
        # for all bash function syntaxes including:
        # - function name() { ... } (with function keyword)
        # - name() { ... } (traditional syntax)
        # - Functions with various indentation levels
        # - Functions with comments before/after/inside

        log.debug(f"Requesting document symbols via LSP for {relative_file_path}")

        # Use the standard LSP approach - bash-language-server handles all function syntaxes correctly
        document_symbols = super().request_document_symbols(relative_file_path, file_buffer=file_buffer)

        # Log detection results for debugging
        functions = [s for s in document_symbols.iter_symbols() if s.get("kind") == 12]
        log.info(f"LSP function detection for {relative_file_path}: Found {len(functions)} functions")

        return document_symbols



"""도메인 예외를 정의한다."""

from dataclasses import dataclass


@dataclass(frozen=True)
class ErrorContext:
    """오류 응답에 필요한 표준 정보를 담는다."""

    code: str
    message: str


class SariBaseError(Exception):
    """모든 도메인 예외의 공통 기반 클래스다."""

    def __init__(self, context: ErrorContext) -> None:
        """예외 생성 시 표준 오류 문맥을 고정한다."""
        super().__init__(context.message)
        self.context = context


class WorkspaceError(SariBaseError):
    """워크스페이스 처리 오류를 표현한다."""


class DaemonError(SariBaseError):
    """데몬 수명주기 오류를 표현한다."""


class ValidationError(SariBaseError):
    """입력 검증 실패를 표현한다."""


class CollectionError(SariBaseError):
    """파일 수집 처리 오류를 표현한다."""


class BenchmarkError(SariBaseError):
    """벤치마크 처리 오류를 표현한다."""


class QualityError(SariBaseError):
    """L3 품질 평가 처리 오류를 표현한다."""



"""LSP 매트릭스 결과 진단 서비스를 제공한다."""

from __future__ import annotations

import json
from pathlib import Path

from sari.core.exceptions import DaemonError, ErrorContext
from sari.core.lsp_provision_policy import get_lsp_provision_policy


class LspMatrixDiagnoseService:
    """LSP 매트릭스 결과를 원인 분해 가능한 진단 리포트로 변환한다."""

    def diagnose(self, matrix_report: dict[str, object]) -> dict[str, object]:
        """매트릭스 리포트를 진단 요약으로 변환한다."""
        languages_raw = matrix_report.get("languages")
        if not isinstance(languages_raw, list):
            raise DaemonError(ErrorContext(code="ERR_LSP_MATRIX_INVALID_RESULT", message="languages must be list"))
        summary_raw = matrix_report.get("summary")
        summary = summary_raw if isinstance(summary_raw, dict) else {}
        gate_raw = matrix_report.get("gate")
        gate = gate_raw if isinstance(gate_raw, dict) else {}
        gate_run_raw = matrix_report.get("gate_run")
        gate_run = gate_run_raw if isinstance(gate_run_raw, dict) else {}

        missing_server_languages: list[str] = []
        timeout_languages: list[str] = []
        symbol_failed_languages: list[str] = []
        error_code_counts: dict[str, int] = {}
        language_policies: list[dict[str, str]] = []

        for item in languages_raw:
            if not isinstance(item, dict):
                continue
            language = str(item.get("language", "")).strip().lower()
            if language == "":
                continue
            error_code = str(item.get("last_error_code", "")).strip()
            if error_code != "":
                error_code_counts[error_code] = error_code_counts.get(error_code, 0) + 1
            policy = get_lsp_provision_policy(language)
            language_policies.append(
                {
                    "language": language,
                    "provisioning_mode": policy.provisioning_mode,
                    "install_hint": policy.install_hint,
                }
            )
            if error_code == "ERR_LSP_SERVER_MISSING":
                missing_server_languages.append(language)
            if error_code == "ERR_LSP_TIMEOUT" or bool(item.get("timeout_occurred")):
                timeout_languages.append(language)
            if not bool(item.get("symbol_extract_success", item.get("available", False))):
                symbol_failed_languages.append(language)

        gate_failed_symbols = gate.get("failed_symbol_languages")
        if isinstance(gate_failed_symbols, list):
            for item in gate_failed_symbols:
                if isinstance(item, str):
                    symbol_failed_languages.append(item.strip().lower())

        diagnosis = {
            "repo_root": str(matrix_report.get("repo_root", "")),
            "run_id": str(matrix_report.get("run_id", "")),
            "gate_decision": str(gate.get("gate_decision", "")),
            "gate_mode": str(gate_run.get("gate_mode", "")),
            "repair_applied": bool(gate_run.get("repair_applied", False)),
            "rerun_count": int(gate_run.get("rerun_count", 0)),
            "final_gate_decision": str(gate_run.get("final_gate_decision", str(gate.get("gate_decision", "")))),
            "readiness_percent": float(summary.get("readiness_percent", 0.0)),
            "symbol_extract_success_rate": float(summary.get("symbol_extract_success_rate", 0.0)),
            "missing_server_languages": sorted(set(missing_server_languages)),
            "timeout_languages": sorted(set(timeout_languages)),
            "symbol_failed_languages": sorted(set(symbol_failed_languages)),
            "error_code_counts": error_code_counts,
            "language_policies": sorted(language_policies, key=lambda item: item["language"]),
            "recommended_actions": self._build_recommended_actions(
                missing_server_languages=sorted(set(missing_server_languages)),
                timeout_languages=sorted(set(timeout_languages)),
                symbol_failed_languages=sorted(set(symbol_failed_languages)),
            ),
        }
        return diagnosis

    def render_markdown(self, diagnosis: dict[str, object]) -> str:
        """진단 JSON을 Markdown 보고서로 렌더링한다."""
        missing = _as_str_list(diagnosis.get("missing_server_languages"))
        timeout = _as_str_list(diagnosis.get("timeout_languages"))
        symbol_failed = _as_str_list(diagnosis.get("symbol_failed_languages"))
        actions = diagnosis.get("recommended_actions")
        action_lines: list[str] = []
        if isinstance(actions, list):
            for item in actions:
                if not isinstance(item, dict):
                    continue
                action_lines.append(
                    f"- [{item.get('severity', 'INFO')}] {item.get('title', '')}: {item.get('message', '')}"
                )
        if len(action_lines) == 0:
            action_lines.append("- [INFO] No action required")

        error_code_counts = diagnosis.get("error_code_counts")
        error_count_lines: list[str] = []
        if isinstance(error_code_counts, dict):
            for code, count in sorted(error_code_counts.items(), key=lambda x: x[0]):
                error_count_lines.append(f"- `{code}`: {int(count)}")
        if len(error_count_lines) == 0:
            error_count_lines.append("- none")
        policy_lines: list[str] = []
        policy_items = diagnosis.get("language_policies")
        if isinstance(policy_items, list):
            for item in policy_items:
                if not isinstance(item, dict):
                    continue
                language = str(item.get("language", "")).strip()
                mode = str(item.get("provisioning_mode", "")).strip()
                hint = str(item.get("install_hint", "")).strip()
                if language == "":
                    continue
                policy_lines.append(f"- `{language}` ({mode}): {hint}")
        if len(policy_lines) == 0:
            policy_lines.append("- none")

        return "\n".join(
            [
                "# LSP Matrix Diagnose Report",
                "",
                f"- repo_root: `{diagnosis.get('repo_root', '')}`",
                f"- run_id: `{diagnosis.get('run_id', '')}`",
                f"- gate_decision: `{diagnosis.get('gate_decision', '')}`",
                f"- gate_mode: `{diagnosis.get('gate_mode', '')}`",
                f"- repair_applied: `{diagnosis.get('repair_applied', False)}`",
                f"- rerun_count: `{diagnosis.get('rerun_count', 0)}`",
                f"- final_gate_decision: `{diagnosis.get('final_gate_decision', '')}`",
                f"- readiness_percent: `{diagnosis.get('readiness_percent', 0.0)}`",
                f"- symbol_extract_success_rate: `{diagnosis.get('symbol_extract_success_rate', 0.0)}`",
                "",
                "## Missing Servers",
                *(["- " + item for item in missing] if len(missing) > 0 else ["- none"]),
                "",
                "## Timeout Languages",
                *(["- " + item for item in timeout] if len(timeout) > 0 else ["- none"]),
                "",
                "## Symbol Failed Languages",
                *(["- " + item for item in symbol_failed] if len(symbol_failed) > 0 else ["- none"]),
                "",
                "## Error Code Counts",
                *error_count_lines,
                "",
                "## Provisioning Policies",
                *policy_lines,
                "",
                "## Recommended Actions",
                *action_lines,
                "",
            ]
        )

    def write_outputs(self, diagnosis: dict[str, object], output_dir: Path) -> tuple[Path, Path]:
        """진단 결과를 JSON/Markdown 파일로 저장한다."""
        output_dir.mkdir(parents=True, exist_ok=True)
        json_path = output_dir / "lsp-matrix-diagnose.json"
        md_path = output_dir / "lsp-matrix-diagnose.md"
        json_path.write_text(json.dumps(diagnosis, ensure_ascii=False, indent=2) + "\n", encoding="utf-8")
        md_path.write_text(self.render_markdown(diagnosis), encoding="utf-8")
        return json_path, md_path

    def _build_recommended_actions(
        self,
        missing_server_languages: list[str],
        timeout_languages: list[str],
        symbol_failed_languages: list[str],
    ) -> list[dict[str, str]]:
        """진단 결과 기반 권장 조치 목록을 생성한다."""
        actions: list[dict[str, str]] = []
        if len(missing_server_languages) > 0:
            hint_fragments: list[str] = []
            for language in missing_server_languages:
                policy = get_lsp_provision_policy(language)
                hint_fragments.append(f"{language}: {policy.install_hint}")
            actions.append(
                {
                    "severity": "HIGH",
                    "title": "Install Missing Language Servers",
                    "message": f"missing server languages: {', '.join(missing_server_languages)}",
                    "recovery_hint": " | ".join(hint_fragments),
                }
            )
        if len(timeout_languages) > 0:
            actions.append(
                {
                    "severity": "MEDIUM",
                    "title": "Review Timeout and Server Health",
                    "message": f"timeout languages: {', '.join(timeout_languages)}",
                }
            )
        if len(symbol_failed_languages) > 0:
            actions.append(
                {
                    "severity": "MEDIUM",
                    "title": "Review DocumentSymbol Extraction",
                    "message": f"symbol failed languages: {', '.join(symbol_failed_languages)}",
                }
            )
        if len(actions) == 0:
            actions.append(
                {
                    "severity": "INFO",
                    "title": "No Immediate Action",
                    "message": "all monitored languages passed current diagnose checks",
                }
            )
        return actions


def _as_str_list(value: object) -> list[str]:
    """값을 문자열 리스트로 정규화한다."""
    if not isinstance(value, list):
        return []
    normalized: list[str] = []
    for item in value:
        if not isinstance(item, str):
            continue
        token = item.strip()
        if token != "":
            normalized.append(token)
    return normalized



"""파일 시스템 이벤트 감시 전용 컴포넌트."""

from __future__ import annotations

import queue
import time
from pathlib import Path
from threading import Event, Lock
from typing import Callable

from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from sari.core.exceptions import CollectionError
from sari.core.models import now_iso8601_utc


class _WatcherHandler(FileSystemEventHandler):
    """watchdog 이벤트를 내부 큐로 전달한다."""

    def __init__(self, event_queue: queue.Queue[tuple[str, str, str]]) -> None:
        """이벤트 전달 큐를 저장한다."""
        super().__init__()
        self._event_queue = event_queue

    def on_created(self, event: FileSystemEvent) -> None:
        """생성 이벤트를 큐에 적재한다."""
        self._enqueue_event("created", event)

    def on_modified(self, event: FileSystemEvent) -> None:
        """수정 이벤트를 큐에 적재한다."""
        self._enqueue_event("modified", event)

    def on_deleted(self, event: FileSystemEvent) -> None:
        """삭제 이벤트를 큐에 적재한다."""
        self._enqueue_event("deleted", event)

    def on_moved(self, event: FileSystemEvent) -> None:
        """이동 이벤트를 큐에 적재한다."""
        self._enqueue_event("moved", event)

    def _enqueue_event(self, event_type: str, event: FileSystemEvent) -> None:
        """디렉터리 이벤트를 제외하고 파일 이벤트를 큐에 적재한다."""
        if event.is_directory:
            return
        self._event_queue.put((event_type, str(event.src_path), str(getattr(event, "dest_path", ""))))


class EventWatcher:
    """watcher 루프/디바운스 처리 책임을 담당한다."""

    def __init__(
        self,
        *,
        workspace_repo: object,
        file_repo: object,
        candidate_index_sink: object | None,
        event_queue: queue.Queue[tuple[str, str, str]],
        stop_event: Event,
        debounce_events: dict[tuple[str, str], tuple[float, str, str]],
        debounce_lock: Lock,
        watcher_debounce_ms: Callable[[], int],
        assert_parent_alive: Callable[[str], None],
        index_file_with_priority: Callable[[str, str, int, str], None],
        handle_background_collection_error: Callable[[CollectionError, str, str], bool],
        priority_high: int,
        set_observer: Callable[[Observer | None], None],
    ) -> None:
        """watcher 동작에 필요한 의존성만 주입받는다."""
        self._workspace_repo = workspace_repo
        self._file_repo = file_repo
        self._candidate_index_sink = candidate_index_sink
        self._event_queue = event_queue
        self._stop_event = stop_event
        self._debounce_events = debounce_events
        self._debounce_lock = debounce_lock
        self._watcher_debounce_ms = watcher_debounce_ms
        self._assert_parent_alive = assert_parent_alive
        self._index_file_with_priority = index_file_with_priority
        self._handle_background_collection_error = handle_background_collection_error
        self._priority_high = priority_high
        self._set_observer = set_observer

    def watcher_loop(self) -> None:
        """watchdog 이벤트 루프를 실행한다."""
        observer = Observer()
        handler = _WatcherHandler(self._event_queue)
        workspaces = self._workspace_repo.list_all()
        for workspace in workspaces:
            if not workspace.is_active:
                continue
            observer.schedule(handler, workspace.path, recursive=True)
        observer.start()
        self._set_observer(observer)
        while not self._stop_event.is_set():
            self._assert_parent_alive(worker_name="watcher")
            self.flush_debounced_events()
            try:
                event_type, src_path, dest_path = self._event_queue.get(timeout=0.5)
            except queue.Empty:
                continue
            self.push_debounced_event(event_type=event_type, src_path=src_path, dest_path=dest_path)

    def handle_fs_event(self, event_type: str, src_path: str, dest_path: str) -> None:
        """단일 파일 시스템 이벤트를 처리한다."""
        source_path = Path(src_path).resolve()
        workspaces = self._workspace_repo.list_all()
        matched_root: Path | None = None
        for workspace in workspaces:
            root = Path(workspace.path).resolve()
            if _path_is_relative_to(source_path, root):
                matched_root = root
                break
        if matched_root is None:
            return
        relative_path = str(source_path.relative_to(matched_root).as_posix())
        if event_type == "deleted":
            self._file_repo.mark_deleted(str(matched_root), relative_path, now_iso8601_utc())
            if self._candidate_index_sink is not None:
                self._candidate_index_sink.record_delete(str(matched_root), relative_path, "watcher_deleted")
            return
        if event_type == "moved" and dest_path.strip() != "":
            self._file_repo.mark_deleted(str(matched_root), relative_path, now_iso8601_utc())
            if self._candidate_index_sink is not None:
                self._candidate_index_sink.record_delete(str(matched_root), relative_path, "watcher_moved")
            moved_dest = Path(dest_path).resolve()
            if _path_is_relative_to(moved_dest, matched_root):
                dest_relative = str(moved_dest.relative_to(matched_root).as_posix())
                try:
                    self._index_file_with_priority(str(matched_root), dest_relative, self._priority_high, "watcher")
                except CollectionError as exc:
                    if self._handle_background_collection_error(exc=exc, phase="watcher_moved", worker_name="watcher"):
                        raise
                    return
            return
        try:
            self._index_file_with_priority(str(matched_root), relative_path, self._priority_high, "watcher")
        except CollectionError as exc:
            if self._handle_background_collection_error(exc=exc, phase="watcher_index", worker_name="watcher"):
                raise
            return

    def push_debounced_event(self, event_type: str, src_path: str, dest_path: str) -> None:
        """디바운스 버퍼에 이벤트를 적재한다."""
        source_path = Path(src_path).resolve()
        workspaces = self._workspace_repo.list_all()
        matched_root: Path | None = None
        for workspace in workspaces:
            root = Path(workspace.path).resolve()
            if _path_is_relative_to(source_path, root):
                matched_root = root
                break
        if matched_root is None:
            return
        relative_path = str(source_path.relative_to(matched_root).as_posix())
        key = (str(matched_root), relative_path)
        now_monotonic = time.monotonic()
        with self._debounce_lock:
            self._debounce_events[key] = (now_monotonic, event_type, dest_path)

    def flush_debounced_events(self) -> None:
        """디바운스 버퍼에서 만료된 이벤트를 처리한다."""
        due_items: list[tuple[str, str, str, str]] = []
        now_monotonic = time.monotonic()
        with self._debounce_lock:
            keys_to_delete: list[tuple[str, str]] = []
            for key, value in self._debounce_events.items():
                ts, event_type, dest_path = value
                if (now_monotonic - ts) * 1000.0 < float(self._watcher_debounce_ms()):
                    continue
                repo_root, relative_path = key
                src_path = str((Path(repo_root) / relative_path).resolve())
                due_items.append((event_type, src_path, dest_path, repo_root))
                keys_to_delete.append(key)
            for key in keys_to_delete:
                self._debounce_events.pop(key, None)
        for event_type, src_path, dest_path, _repo_root in due_items:
            self.handle_fs_event(event_type=event_type, src_path=src_path, dest_path=dest_path)


def _path_is_relative_to(path: Path, base: Path) -> bool:
    """path가 base 하위인지 판정한다."""
    try:
        path.relative_to(base)
    except ValueError:
        return False
    return True



"""
Provides Ruby specific instantiation of the LanguageServer class using Solargraph.
Contains various configurations and settings specific to Ruby.
"""

import json
import logging
import os
import pathlib
import re
import shutil
import subprocess
import threading

from overrides import override

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class Solargraph(SolidLanguageServer):
    """
    Provides Ruby specific instantiation of the LanguageServer class using Solargraph.
    Contains various configurations and settings specific to Ruby.
    """

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a Solargraph instance. This class is not meant to be instantiated directly.
        Use LanguageServer.create() instead.
        """
        solargraph_executable_path = self._setup_runtime_dependencies(config, repository_root_path)
        super().__init__(
            config,
            repository_root_path,
            ProcessLaunchInfo(cmd=f"{solargraph_executable_path} stdio", cwd=repository_root_path),
            "ruby",
            solidlsp_settings,
        )
        # Override internal language enum for file matching (excludes .erb files)
        # while keeping LSP languageId as "ruby" for protocol compliance
        from solidlsp.ls_config import Language

        self.language = Language.RUBY_SOLARGRAPH
        self.analysis_complete = threading.Event()
        self.service_ready_event = threading.Event()
        self.initialize_searcher_command_available = threading.Event()
        self.resolve_main_method_available = threading.Event()

        # Set timeout for Solargraph requests - Bundler environments may need more time
        self.set_request_timeout(120.0)  # 120 seconds for initialization and requests

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        ruby_ignored_dirs = [
            "vendor",  # Ruby vendor directory
            ".bundle",  # Bundler cache
            "tmp",  # Temporary files
            "log",  # Log files
            "coverage",  # Test coverage reports
            ".yardoc",  # YARD documentation cache
            "doc",  # Generated documentation
            "node_modules",  # Node modules (for Rails with JS)
            "storage",  # Active Storage files (Rails)
        ]
        return super().is_ignored_dirname(dirname) or dirname in ruby_ignored_dirs

    @staticmethod
    def _setup_runtime_dependencies(config: LanguageServerConfig, repository_root_path: str) -> str:
        """
        Setup runtime dependencies for Solargraph and return the command to start the server.
        """
        # Check if Ruby is installed
        try:
            result = subprocess.run(["ruby", "--version"], check=True, capture_output=True, cwd=repository_root_path, text=True)
            ruby_version = result.stdout.strip()
            log.info(f"Ruby version: {ruby_version}")

            # Extract version number for compatibility checks
            version_match = re.search(r"ruby (\d+)\.(\d+)\.(\d+)", ruby_version)
            if version_match:
                major, minor, patch = map(int, version_match.groups())
                if major < 2 or (major == 2 and minor < 6):
                    log.warning(f"Warning: Ruby {major}.{minor}.{patch} detected. Solargraph works best with Ruby 2.6+")

        except subprocess.CalledProcessError as e:
            error_msg = e.stderr.decode() if e.stderr else "Unknown error"
            raise RuntimeError(
                f"Error checking Ruby installation: {error_msg}. Please ensure Ruby is properly installed and in PATH."
            ) from e
        except FileNotFoundError as e:
            raise RuntimeError(
                "Ruby is not installed or not found in PATH. Please install Ruby using one of these methods:\n"
                "  - Using rbenv: rbenv install 3.0.0 && rbenv global 3.0.0\n"
                "  - Using RVM: rvm install 3.0.0 && rvm use 3.0.0 --default\n"
                "  - Using asdf: asdf install ruby 3.0.0 && asdf global ruby 3.0.0\n"
                "  - System package manager (brew install ruby, apt install ruby, etc.)"
            ) from e

        # Helper function for Windows-compatible executable search
        def find_executable_with_extensions(executable_name: str) -> str | None:
            """Find executable with Windows-specific extensions if on Windows."""
            import platform

            if platform.system() == "Windows":
                for ext in [".bat", ".cmd", ".exe"]:
                    path = shutil.which(f"{executable_name}{ext}")
                    if path:
                        return path
                return shutil.which(executable_name)
            else:
                return shutil.which(executable_name)

        # Check for Bundler project (Gemfile exists)
        gemfile_path = os.path.join(repository_root_path, "Gemfile")
        gemfile_lock_path = os.path.join(repository_root_path, "Gemfile.lock")
        is_bundler_project = os.path.exists(gemfile_path)

        if is_bundler_project:
            log.info("Detected Bundler project (Gemfile found)")

            # Check if bundle command is available
            bundle_path = find_executable_with_extensions("bundle")
            if not bundle_path:
                # Try common bundle executables
                for bundle_cmd in ["bin/bundle", "bundle"]:
                    if bundle_cmd.startswith("bin/"):
                        bundle_full_path = os.path.join(repository_root_path, bundle_cmd)
                    else:
                        bundle_full_path = find_executable_with_extensions(bundle_cmd)  # type: ignore[assignment]
                    if bundle_full_path and os.path.exists(bundle_full_path):
                        bundle_path = bundle_full_path if bundle_cmd.startswith("bin/") else bundle_cmd
                        break

            if not bundle_path:
                raise RuntimeError(
                    "Bundler project detected but 'bundle' command not found. Please install Bundler:\n"
                    "  - gem install bundler\n"
                    "  - Or use your Ruby version manager's bundler installation\n"
                    "  - Ensure the bundle command is in your PATH"
                )

            # Check if solargraph is in Gemfile.lock
            solargraph_in_bundle = False
            if os.path.exists(gemfile_lock_path):
                try:
                    with open(gemfile_lock_path) as f:
                        content = f.read()
                        solargraph_in_bundle = "solargraph" in content.lower()
                except OSError as e:
                    log.warning(f"Warning: Could not read Gemfile.lock: {e}")

            if solargraph_in_bundle:
                log.info("Found solargraph in Gemfile.lock")
                return f"{bundle_path} exec solargraph"
            else:
                log.warning(
                    "solargraph not found in Gemfile.lock. Please add 'gem \"solargraph\"' to your Gemfile and run 'bundle install'",
                )
                # Fall through to global installation check

        # Check if solargraph is installed globally
        # First, try to find solargraph in PATH (includes asdf shims) with Windows support
        solargraph_path = find_executable_with_extensions("solargraph")
        if solargraph_path:
            log.info(f"Found solargraph at: {solargraph_path}")
            return solargraph_path

        # Fallback to gem exec (for non-Bundler projects or when global solargraph not found)
        if not is_bundler_project:
            runtime_dependencies = [
                {
                    "url": "https://rubygems.org/downloads/solargraph-0.51.1.gem",
                    "installCommand": "gem install solargraph -v 0.51.1",
                    "binaryName": "solargraph",
                    "archiveType": "gem",
                }
            ]

            dependency = runtime_dependencies[0]
            try:
                result = subprocess.run(
                    ["gem", "list", "^solargraph$", "-i"], check=False, capture_output=True, text=True, cwd=repository_root_path
                )
                if result.stdout.strip() == "false":
                    log.info("Installing Solargraph...")
                    subprocess.run(dependency["installCommand"].split(), check=True, capture_output=True, cwd=repository_root_path)

                return "gem exec solargraph"
            except subprocess.CalledProcessError as e:
                error_msg = e.stderr.decode() if e.stderr else str(e)
                raise RuntimeError(
                    f"Failed to check or install Solargraph: {error_msg}\nPlease try installing manually: gem install solargraph"
                ) from e
        else:
            raise RuntimeError(
                "This appears to be a Bundler project, but solargraph is not available. "
                "Please add 'gem \"solargraph\"' to your Gemfile and run 'bundle install'."
            )

    @staticmethod
    def _detect_rails_project(repository_root_path: str) -> bool:
        """
        Detect if this is a Rails project by checking for Rails-specific files.
        """
        rails_indicators = [
            "config/application.rb",
            "config/environment.rb",
            "app/controllers/application_controller.rb",
            "Rakefile",
        ]

        for indicator in rails_indicators:
            if os.path.exists(os.path.join(repository_root_path, indicator)):
                return True

        # Check for Rails in Gemfile
        gemfile_path = os.path.join(repository_root_path, "Gemfile")
        if os.path.exists(gemfile_path):
            try:
                with open(gemfile_path) as f:
                    content = f.read().lower()
                    if "gem 'rails'" in content or 'gem "rails"' in content:
                        return True
            except OSError as exc:
                log.debug(f"Failed to read Gemfile for Rails detection: {exc}")

        return False

    @staticmethod
    def _get_ruby_exclude_patterns(repository_root_path: str) -> list[str]:
        """
        Get Ruby and Rails-specific exclude patterns for better performance.
        """
        base_patterns = [
            "**/vendor/**",  # Ruby vendor directory (similar to node_modules)
            "**/.bundle/**",  # Bundler cache
            "**/tmp/**",  # Temporary files
            "**/log/**",  # Log files
            "**/coverage/**",  # Test coverage reports
            "**/.yardoc/**",  # YARD documentation cache
            "**/doc/**",  # Generated documentation
            "**/.git/**",  # Git directory
            "**/node_modules/**",  # Node modules (for Rails with JS)
            "**/public/assets/**",  # Rails compiled assets
        ]

        # Add Rails-specific patterns if this is a Rails project
        if Solargraph._detect_rails_project(repository_root_path):
            rails_patterns = [
                "**/public/packs/**",  # Webpacker output
                "**/public/webpack/**",  # Webpack output
                "**/storage/**",  # Active Storage files
                "**/tmp/cache/**",  # Rails cache
                "**/tmp/pids/**",  # Process IDs
                "**/tmp/sessions/**",  # Session files
                "**/tmp/sockets/**",  # Socket files
                "**/db/*.sqlite3",  # SQLite databases
            ]
            base_patterns.extend(rails_patterns)

        return base_patterns

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Solargraph Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        exclude_patterns = Solargraph._get_ruby_exclude_patterns(repository_absolute_path)

        initialize_params: InitializeParams = {  # type: ignore
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "initializationOptions": {
                "exclude": exclude_patterns,  # type: ignore[dict-item]
            },
            "capabilities": {
                "workspace": {
                    "workspaceEdit": {"documentChanges": True},
                },
                "textDocument": {
                    "documentSymbol": {
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},  # type: ignore[arg-type]
                    },
                },
            },
            "trace": "verbose",  # type: ignore[typeddict-item]
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }
        return initialize_params  # type: ignore[return-value]

    def _start_server(self) -> None:
        """
        Starts the Solargraph Language Server for Ruby
        """

        def register_capability_handler(params: dict) -> None:
            assert "registrations" in params
            for registration in params["registrations"]:
                if registration["method"] == "workspace/executeCommand":
                    self.initialize_searcher_command_available.set()
                    self.resolve_main_method_available.set()
            return

        def lang_status_handler(params: dict) -> None:
            log.info(f"LSP: language/status: {params}")
            if params.get("type") == "ServiceReady" and params.get("message") == "Service is ready.":
                log.info("Solargraph service is ready.")
                self.analysis_complete.set()

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("language/status", lang_status_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)
        self.server.on_notification("language/actionableNotification", do_nothing)

        log.info("Starting solargraph server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        log.info(f"Sending init params: {json.dumps(initialize_params, indent=4)}")
        init_response = self.server.send.initialize(initialize_params)
        log.info(f"Received init response: {init_response}")
        assert init_response["capabilities"]["textDocumentSync"] == 2
        assert "completionProvider" in init_response["capabilities"]
        assert init_response["capabilities"]["completionProvider"] == {
            "resolveProvider": True,
            "triggerCharacters": [".", ":", "@"],
        }
        self.server.notify.initialized({})

        # Wait for Solargraph to complete its initial workspace analysis
        # This prevents issues by ensuring background tasks finish
        log.info("Waiting for Solargraph to complete initial workspace analysis...")
        if self.analysis_complete.wait(timeout=60.0):
            log.info("Solargraph initial analysis complete, server ready")
        else:
            log.warning("Timeout waiting for Solargraph analysis completion, proceeding anyway")
            # Fallback: assume analysis is complete after timeout
            self.analysis_complete.set()



"""파이프라인 품질 실행 저장소를 구현한다."""

from __future__ import annotations

import json
import uuid
from pathlib import Path

from sari.db.row_mapper import row_int, row_optional_str, row_str
from sari.db.schema import connect


class PipelineQualityRepository:
    """품질 실행 이력을 영속화한다."""

    def __init__(self, db_path: Path) -> None:
        """저장소 DB 경로를 저장한다."""
        self._db_path = db_path

    def create_run(self, repo_root: str, limit_files: int, profile: str, started_at: str) -> str:
        """신규 품질 실행을 생성한다."""
        run_id = str(uuid.uuid4())
        with connect(self._db_path) as conn:
            conn.execute(
                """
                INSERT INTO pipeline_quality_runs(
                    run_id, repo_root, limit_files, profile, started_at, finished_at, status, summary_json
                )
                VALUES(
                    :run_id, :repo_root, :limit_files, :profile, :started_at, NULL, 'RUNNING', NULL
                )
                """,
                {
                    "run_id": run_id,
                    "repo_root": repo_root,
                    "limit_files": limit_files,
                    "profile": profile,
                    "started_at": started_at,
                },
            )
            conn.commit()
        return run_id

    def complete_run(self, run_id: str, finished_at: str, status: str, summary: dict[str, object]) -> None:
        """품질 실행을 완료 상태로 갱신한다."""
        with connect(self._db_path) as conn:
            conn.execute(
                """
                UPDATE pipeline_quality_runs
                SET finished_at = :finished_at,
                    status = :status,
                    summary_json = :summary_json
                WHERE run_id = :run_id
                """,
                {
                    "run_id": run_id,
                    "finished_at": finished_at,
                    "status": status,
                    "summary_json": json.dumps(summary, ensure_ascii=False),
                },
            )
            conn.commit()

    def get_latest_run(self) -> dict[str, object] | None:
        """최신 품질 실행 결과를 반환한다."""
        with connect(self._db_path) as conn:
            row = conn.execute(
                """
                SELECT run_id, repo_root, limit_files, profile, started_at, finished_at, status, summary_json
                FROM pipeline_quality_runs
                ORDER BY started_at DESC
                LIMIT 1
                """
            ).fetchone()
        if row is None:
            return None

        summary_raw = row_optional_str(row, "summary_json")
        summary: dict[str, object] = {}
        if summary_raw is not None:
            parsed = json.loads(summary_raw)
            if isinstance(parsed, dict):
                summary = parsed

        return {
            "run_id": row_str(row, "run_id"),
            "repo_root": row_str(row, "repo_root"),
            "limit_files": row_int(row, "limit_files"),
            "profile": row_str(row, "profile"),
            "started_at": row_str(row, "started_at"),
            "finished_at": row_optional_str(row, "finished_at"),
            "status": row_str(row, "status"),
            "summary": summary,
        }



"""
Provides Clojure specific instantiation of the LanguageServer class. Contains various configurations and settings specific to Clojure.
"""

import logging
import os
import pathlib
import shutil
import subprocess
import threading
from typing import cast

from solidlsp.ls import LanguageServerDependencyProvider, LanguageServerDependencyProviderSinglePath, SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.settings import SolidLSPSettings

from .common import RuntimeDependency, RuntimeDependencyCollection

log = logging.getLogger(__name__)


def run_command(cmd: list, capture_output: bool = True) -> subprocess.CompletedProcess:
    return subprocess.run(
        cmd, stdout=subprocess.PIPE if capture_output else None, stderr=subprocess.STDOUT if capture_output else None, text=True, check=True
    )


def verify_clojure_cli() -> None:
    install_msg = "Please install the official Clojure CLI from:\n  https://clojure.org/guides/getting_started"
    if shutil.which("clojure") is None:
        raise FileNotFoundError("`clojure` not found.\n" + install_msg)

    help_proc = run_command(["clojure", "--help"])
    if "-Aaliases" not in help_proc.stdout:
        raise RuntimeError("Detected a Clojure executable, but it does not support '-Aaliases'.\n" + install_msg)

    spath_proc = run_command(["clojure", "-Spath"], capture_output=False)
    if spath_proc.returncode != 0:
        raise RuntimeError("`clojure -Spath` failed; please upgrade to Clojure CLI ≥ 1.10.")


class ClojureLSP(SolidLanguageServer):
    """
    Provides a clojure-lsp specific instantiation of the LanguageServer class. Contains various configurations and settings specific to clojure.
    """

    clojure_lsp_releases = "https://github.com/clojure-lsp/clojure-lsp/releases/latest/download"
    runtime_dependencies = RuntimeDependencyCollection(
        [
            RuntimeDependency(
                id="clojure-lsp",
                url=f"{clojure_lsp_releases}/clojure-lsp-native-macos-aarch64.zip",
                platform_id="osx-arm64",
                archive_type="zip",
                binary_name="clojure-lsp",
            ),
            RuntimeDependency(
                id="clojure-lsp",
                url=f"{clojure_lsp_releases}/clojure-lsp-native-macos-amd64.zip",
                platform_id="osx-x64",
                archive_type="zip",
                binary_name="clojure-lsp",
            ),
            RuntimeDependency(
                id="clojure-lsp",
                url=f"{clojure_lsp_releases}/clojure-lsp-native-linux-aarch64.zip",
                platform_id="linux-arm64",
                archive_type="zip",
                binary_name="clojure-lsp",
            ),
            RuntimeDependency(
                id="clojure-lsp",
                url=f"{clojure_lsp_releases}/clojure-lsp-native-linux-amd64.zip",
                platform_id="linux-x64",
                archive_type="zip",
                binary_name="clojure-lsp",
            ),
            RuntimeDependency(
                id="clojure-lsp",
                url=f"{clojure_lsp_releases}/clojure-lsp-native-windows-amd64.zip",
                platform_id="win-x64",
                archive_type="zip",
                binary_name="clojure-lsp.exe",
            ),
        ]
    )

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        """
        Creates a ClojureLSP instance. This class is not meant to be instantiated directly. Use LanguageServer.create() instead.
        """
        super().__init__(
            config,
            repository_root_path,
            None,
            "clojure",
            solidlsp_settings,
        )
        self.server_ready = threading.Event()
        self.initialize_searcher_command_available = threading.Event()
        self.resolve_main_method_available = threading.Event()
        self.service_ready_event = threading.Event()

    def _create_dependency_provider(self) -> LanguageServerDependencyProvider:
        return self.DependencyProvider(self._custom_settings, self._ls_resources_dir)

    class DependencyProvider(LanguageServerDependencyProviderSinglePath):
        def _get_or_install_core_dependency(self) -> str:
            """Setup runtime dependencies for clojure-lsp and return the path to the executable."""
            verify_clojure_cli()
            deps = ClojureLSP.runtime_dependencies
            dependency = deps.get_single_dep_for_current_platform()

            clojurelsp_executable_path = deps.binary_path(self._ls_resources_dir)
            if not os.path.exists(clojurelsp_executable_path):
                log.info(
                    f"Downloading and extracting clojure-lsp from {dependency.url} to {self._ls_resources_dir}",
                )
                deps.install(self._ls_resources_dir)
            if not os.path.exists(clojurelsp_executable_path):
                raise FileNotFoundError(f"Download failed? Could not find clojure-lsp executable at {clojurelsp_executable_path}")
            os.chmod(clojurelsp_executable_path, 0o755)
            return clojurelsp_executable_path

        def _create_launch_command(self, core_path: str) -> list[str] | str:
            return [core_path]

    @staticmethod
    def _get_initialize_params(repository_absolute_path: str) -> InitializeParams:
        """Returns the init params for clojure-lsp."""
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        result = {  # type: ignore
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "capabilities": {
                "workspace": {
                    "applyEdit": True,
                    "workspaceEdit": {"documentChanges": True},
                    "symbol": {"symbolKind": {"valueSet": list(range(1, 27))}},
                    "workspaceFolders": True,
                },
                "textDocument": {
                    "synchronization": {"didSave": True},
                    "publishDiagnostics": {"relatedInformation": True, "tagSupport": {"valueSet": [1, 2]}},
                    "definition": {"linkSupport": True},
                    "references": {},
                    "hover": {"contentFormat": ["markdown", "plaintext"]},
                    "documentSymbol": {
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},  #
                    },
                },
                "general": {"positionEncodings": ["utf-16"]},
            },
            "initializationOptions": {"dependency-scheme": "jar", "text-document-sync-kind": "incremental"},
            "trace": "off",
            "workspaceFolders": [{"uri": root_uri, "name": os.path.basename(repository_absolute_path)}],
        }
        return cast(InitializeParams, result)

    def _start_server(self) -> None:
        def register_capability_handler(params: dict) -> None:
            assert "registrations" in params
            for registration in params["registrations"]:
                if registration["method"] == "workspace/executeCommand":
                    self.initialize_searcher_command_available.set()
                    self.resolve_main_method_available.set()
            return

        def lang_status_handler(params: dict) -> None:
            # 참고: 필요 시 아래 상태 알림까지 대기하도록 확장 가능
            # server -> client: {'jsonrpc': '2.0', 'method': 'language/status', 'params': {'type': 'ProjectStatus', 'message': 'OK'}}
            # Before proceeding?
            if params["type"] == "ServiceReady" and params["message"] == "ServiceReady":
                self.service_ready_event.set()

        def execute_client_command_handler(params: dict) -> list:
            return []

        def do_nothing(params: dict) -> None:
            return

        def check_experimental_status(params: dict) -> None:
            if params["quiescent"] is True:
                self.server_ready.set()

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("language/status", lang_status_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_request("workspace/executeClientCommand", execute_client_command_handler)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)
        self.server.on_notification("language/actionableNotification", do_nothing)
        self.server.on_notification("experimental/serverStatus", check_experimental_status)

        log.info("Starting clojure-lsp server process")
        self.server.start()

        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)
        assert init_response["capabilities"]["textDocumentSync"]["change"] == 2  # type: ignore
        assert "completionProvider" in init_response["capabilities"]
        # Clojure-lsp completion provider capabilities are more flexible than other servers'
        completion_provider = init_response["capabilities"]["completionProvider"]
        assert completion_provider["resolveProvider"] is True
        assert "triggerCharacters" in completion_provider
        self.server.notify.initialized({})
        # after initialize, Clojure-lsp is ready to serve
        self.server_ready.set()



from __future__ import annotations

from enum import Enum, IntEnum, IntFlag
from typing import Literal, NotRequired, Union

from typing_extensions import TypedDict

class NotebookDocumentSyncRegistrationOptions(TypedDict):
    """Registration options specific to a notebook.

    @since 3.17.0
    """

    notebookSelector: list[
        Union[
            "__NotebookDocumentSyncOptions_notebookSelector_Type_3",
            "__NotebookDocumentSyncOptions_notebookSelector_Type_4",
        ]
    ]
    """ The notebooks to be synced """
    save: NotRequired[bool]
    """ Whether save notification should be forwarded to
    the server. Will only be honored if mode === `notebook`. """
    id: NotRequired[str]
    """ The id used to register the request. The id can be used to deregister
    the request again. See also Registration#id. """


class WorkspaceFoldersServerCapabilities(TypedDict):
    supported: NotRequired[bool]
    """ The server has support for workspace folders """
    changeNotifications: NotRequired[str | bool]
    """ Whether the server wants to receive workspace folder
    change notifications.

    If a string is provided the string is treated as an ID
    under which the notification is registered on the client
    side. The ID can be used to unregister for these events
    using the `client/unregisterCapability` request. """


class FileOperationOptions(TypedDict):
    """Options for notifications/requests for user operations on files.

    @since 3.16.0
    """

    didCreate: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving didCreateFiles notifications. """
    willCreate: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving willCreateFiles requests. """
    didRename: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving didRenameFiles notifications. """
    willRename: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving willRenameFiles requests. """
    didDelete: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving didDeleteFiles file notifications. """
    willDelete: NotRequired["FileOperationRegistrationOptions"]
    """ The server is interested in receiving willDeleteFiles file requests. """


class CodeDescription(TypedDict):
    """Structure to capture a description for an error code.

    @since 3.16.0
    """

    href: "URI"
    """ An URI to open with more information about the diagnostic error. """


class DiagnosticRelatedInformation(TypedDict):
    """Represents a related message and source code location for a diagnostic. This should be
    used to point to code locations that cause or related to a diagnostics, e.g when duplicating
    a symbol in a scope.
    """

    location: "Location"
    """ The location of this related diagnostic information. """
    message: str
    """ The message of this related diagnostic information. """


class ParameterInformation(TypedDict):
    """Represents a parameter of a callable-signature. A parameter can
    have a label and a doc-comment.
    """

    label: str | list[Uint | Uint]
    """ The label of this parameter information.

    Either a string or an inclusive start and exclusive end offsets within its containing
    signature label. (see SignatureInformation.label). The offsets are based on a UTF-16
    string representation as `Position` and `Range` does.

    *Note*: a label of type string should be a substring of its containing signature label.
    Its intended use case is to highlight the parameter label part in the `SignatureInformation.label`. """
    documentation: NotRequired[Union[str, "MarkupContent"]]
    """ The human-readable doc-comment of this parameter. Will be shown
    in the UI but can be omitted. """


class NotebookCellTextDocumentFilter(TypedDict):
    """A notebook cell text document filter denotes a cell text
    document by different properties.

    @since 3.17.0
    """

    notebook: Union[str, "NotebookDocumentFilter"]
    """ A filter that matches against the notebook
    containing the notebook cell. If a string
    value is provided it matches against the
    notebook type. '*' matches every notebook. """
    language: NotRequired[str]
    """ A language id like `python`.

    Will be matched against the language id of the
    notebook cell document. '*' matches every language. """


class FileOperationPatternOptions(TypedDict):
    """Matching options for the file operation pattern.

    @since 3.16.0
    """

    ignoreCase: NotRequired[bool]
    """ The pattern should be matched ignoring casing. """


class ExecutionSummary(TypedDict):
    executionOrder: Uint
    """ A strict monotonically increasing value
    indicating the execution order of a cell
    inside a notebook. """
    success: NotRequired[bool]
    """ Whether the execution was successful or
    not if known by the client. """


class WorkspaceClientCapabilities(TypedDict):
    """Workspace specific client capabilities."""

    applyEdit: NotRequired[bool]
    """ The client supports applying batch edits
    to the workspace by supporting the request
    'workspace/applyEdit' """
    workspaceEdit: NotRequired["WorkspaceEditClientCapabilities"]
    """ Capabilities specific to `WorkspaceEdit`s. """
    didChangeConfiguration: NotRequired["DidChangeConfigurationClientCapabilities"]
    """ Capabilities specific to the `workspace/didChangeConfiguration` notification. """
    didChangeWatchedFiles: NotRequired["DidChangeWatchedFilesClientCapabilities"]
    """ Capabilities specific to the `workspace/didChangeWatchedFiles` notification. """
    symbol: NotRequired["WorkspaceSymbolClientCapabilities"]
    """ Capabilities specific to the `workspace/symbol` request. """
    executeCommand: NotRequired["ExecuteCommandClientCapabilities"]
    """ Capabilities specific to the `workspace/executeCommand` request. """
    workspaceFolders: NotRequired[bool]
    """ The client has support for workspace folders.

    @since 3.6.0 """
    configuration: NotRequired[bool]
    """ The client supports `workspace/configuration` requests.

    @since 3.6.0 """
    semanticTokens: NotRequired["SemanticTokensWorkspaceClientCapabilities"]
    """ Capabilities specific to the semantic token requests scoped to the
    workspace.

    @since 3.16.0. """
    codeLens: NotRequired["CodeLensWorkspaceClientCapabilities"]
    """ Capabilities specific to the code lens requests scoped to the
    workspace.

    @since 3.16.0. """
    fileOperations: NotRequired["FileOperationClientCapabilities"]
    """ The client has support for file notifications/requests for user operations on files.

    Since 3.16.0 """
    inlineValue: NotRequired["InlineValueWorkspaceClientCapabilities"]
    """ Capabilities specific to the inline values requests scoped to the
    workspace.

    @since 3.17.0. """
    inlayHint: NotRequired["InlayHintWorkspaceClientCapabilities"]
    """ Capabilities specific to the inlay hint requests scoped to the
    workspace.

    @since 3.17.0. """
    diagnostics: NotRequired["DiagnosticWorkspaceClientCapabilities"]
    """ Capabilities specific to the diagnostic requests scoped to the
    workspace.

    @since 3.17.0. """


class TextDocumentClientCapabilities(TypedDict):
    """Text document specific client capabilities."""

    synchronization: NotRequired["TextDocumentSyncClientCapabilities"]
    """ Defines which synchronization capabilities the client supports. """
    completion: NotRequired["CompletionClientCapabilities"]
    """ Capabilities specific to the `textDocument/completion` request. """
    hover: NotRequired["HoverClientCapabilities"]
    """ Capabilities specific to the `textDocument/hover` request. """
    signatureHelp: NotRequired["SignatureHelpClientCapabilities"]
    """ Capabilities specific to the `textDocument/signatureHelp` request. """
    declaration: NotRequired["DeclarationClientCapabilities"]
    """ Capabilities specific to the `textDocument/declaration` request.

    @since 3.14.0 """
    definition: NotRequired["DefinitionClientCapabilities"]
    """ Capabilities specific to the `textDocument/definition` request. """
    typeDefinition: NotRequired["TypeDefinitionClientCapabilities"]
    """ Capabilities specific to the `textDocument/typeDefinition` request.

    @since 3.6.0 """
    implementation: NotRequired["ImplementationClientCapabilities"]
    """ Capabilities specific to the `textDocument/implementation` request.

    @since 3.6.0 """
    references: NotRequired["ReferenceClientCapabilities"]
    """ Capabilities specific to the `textDocument/references` request. """
    documentHighlight: NotRequired["DocumentHighlightClientCapabilities"]
    """ Capabilities specific to the `textDocument/documentHighlight` request. """
    documentSymbol: NotRequired["DocumentSymbolClientCapabilities"]
    """ Capabilities specific to the `textDocument/documentSymbol` request. """
    codeAction: NotRequired["CodeActionClientCapabilities"]
    """ Capabilities specific to the `textDocument/codeAction` request. """
    codeLens: NotRequired["CodeLensClientCapabilities"]
    """ Capabilities specific to the `textDocument/codeLens` request. """
    documentLink: NotRequired["DocumentLinkClientCapabilities"]
    """ Capabilities specific to the `textDocument/documentLink` request. """
    colorProvider: NotRequired["DocumentColorClientCapabilities"]
    """ Capabilities specific to the `textDocument/documentColor` and the
    `textDocument/colorPresentation` request.

    @since 3.6.0 """
    formatting: NotRequired["DocumentFormattingClientCapabilities"]
    """ Capabilities specific to the `textDocument/formatting` request. """
    rangeFormatting: NotRequired["DocumentRangeFormattingClientCapabilities"]
    """ Capabilities specific to the `textDocument/rangeFormatting` request. """
    onTypeFormatting: NotRequired["DocumentOnTypeFormattingClientCapabilities"]
    """ Capabilities specific to the `textDocument/onTypeFormatting` request. """
    rename: NotRequired["RenameClientCapabilities"]
    """ Capabilities specific to the `textDocument/rename` request. """
    foldingRange: NotRequired["FoldingRangeClientCapabilities"]
    """ Capabilities specific to the `textDocument/foldingRange` request.

    @since 3.10.0 """
    selectionRange: NotRequired["SelectionRangeClientCapabilities"]
    """ Capabilities specific to the `textDocument/selectionRange` request.

    @since 3.15.0 """
    publishDiagnostics: NotRequired["PublishDiagnosticsClientCapabilities"]
    """ Capabilities specific to the `textDocument/publishDiagnostics` notification. """
    callHierarchy: NotRequired["CallHierarchyClientCapabilities"]
    """ Capabilities specific to the various call hierarchy requests.

    @since 3.16.0 """
    semanticTokens: NotRequired["SemanticTokensClientCapabilities"]
    """ Capabilities specific to the various semantic token request.

    @since 3.16.0 """
    linkedEditingRange: NotRequired["LinkedEditingRangeClientCapabilities"]
    """ Capabilities specific to the `textDocument/linkedEditingRange` request.

    @since 3.16.0 """
    moniker: NotRequired["MonikerClientCapabilities"]
    """ Client capabilities specific to the `textDocument/moniker` request.

    @since 3.16.0 """
    typeHierarchy: NotRequired["TypeHierarchyClientCapabilities"]
    """ Capabilities specific to the various type hierarchy requests.

    @since 3.17.0 """
    inlineValue: NotRequired["InlineValueClientCapabilities"]
    """ Capabilities specific to the `textDocument/inlineValue` request.

    @since 3.17.0 """
    inlayHint: NotRequired["InlayHintClientCapabilities"]
    """ Capabilities specific to the `textDocument/inlayHint` request.

    @since 3.17.0 """
    diagnostic: NotRequired["DiagnosticClientCapabilities"]
    """ Capabilities specific to the diagnostic pull model.

    @since 3.17.0 """


class NotebookDocumentClientCapabilities(TypedDict):
    """Capabilities specific to the notebook document support.

    @since 3.17.0
    """

    synchronization: "NotebookDocumentSyncClientCapabilities"
    """ Capabilities specific to notebook document synchronization

    @since 3.17.0 """


class WindowClientCapabilities(TypedDict):
    workDoneProgress: NotRequired[bool]
    """ It indicates whether the client supports server initiated
    progress using the `window/workDoneProgress/create` request.

    The capability also controls Whether client supports handling
    of progress notifications. If set servers are allowed to report a
    `workDoneProgress` property in the request specific server
    capabilities.

    @since 3.15.0 """
    showMessage: NotRequired["ShowMessageRequestClientCapabilities"]
    """ Capabilities specific to the showMessage request.

    @since 3.16.0 """
    showDocument: NotRequired["ShowDocumentClientCapabilities"]
    """ Capabilities specific to the showDocument request.

    @since 3.16.0 """


class GeneralClientCapabilities(TypedDict):
    """General client capabilities.

    @since 3.16.0
    """

    staleRequestSupport: NotRequired["__GeneralClientCapabilities_staleRequestSupport_Type_1"]
    """ Client capability that signals how the client
    handles stale requests (e.g. a request
    for which the client will not process the response
    anymore since the information is outdated).

    @since 3.17.0 """
    regularExpressions: NotRequired["RegularExpressionsClientCapabilities"]
    """ Client capabilities specific to regular expressions.

    @since 3.16.0 """
    markdown: NotRequired["MarkdownClientCapabilities"]
    """ Client capabilities specific to the client's markdown parser.

    @since 3.16.0 """
    positionEncodings: NotRequired[list["PositionEncodingKind"]]
    """ The position encodings supported by the client. Client and server
    have to agree on the same position encoding to ensure that offsets
    (e.g. character position in a line) are interpreted the same on both
    sides.

    To keep the protocol backwards compatible the following applies: if
    the value 'utf-16' is missing from the array of position encodings
    servers can assume that the client supports UTF-16. UTF-16 is
    therefore a mandatory encoding.

    If omitted it defaults to ['utf-16'].

    Implementation considerations: since the conversion from one encoding
    into another requires the content of the file / line the conversion
    is best done where the file is read which is usually on the server
    side.

    @since 3.17.0 """


class RelativePattern(TypedDict):
    """A relative pattern is a helper to construct glob patterns that are matched
    relatively to a base URI. The common value for a `baseUri` is a workspace
    folder root, but it can be another absolute URI as well.

    @since 3.17.0
    """

    baseUri: Union["WorkspaceFolder", "URI"]
    """ A workspace folder or a base URI to which this pattern will be matched
    against relatively. """
    pattern: "Pattern"
    """ The actual glob pattern; """


class WorkspaceEditClientCapabilities(TypedDict):
    documentChanges: NotRequired[bool]
    """ The client supports versioned document changes in `WorkspaceEdit`s """
    resourceOperations: NotRequired[list["ResourceOperationKind"]]
    """ The resource operations the client supports. Clients should at least
    support 'create', 'rename' and 'delete' files and folders.

    @since 3.13.0 """
    failureHandling: NotRequired["FailureHandlingKind"]
    """ The failure handling strategy of a client if applying the workspace edit
    fails.

    @since 3.13.0 """
    normalizesLineEndings: NotRequired[bool]
    """ Whether the client normalizes line endings to the client specific
    setting.
    If set to `true` the client will normalize line ending characters
    in a workspace edit to the client-specified new line
    character.

    @since 3.16.0 """
    changeAnnotationSupport: NotRequired["__WorkspaceEditClientCapabilities_changeAnnotationSupport_Type_1"]
    """ Whether the client in general supports change annotations on text edits,
    create file, rename file and delete file changes.

    @since 3.16.0 """


class DidChangeConfigurationClientCapabilities(TypedDict):
    dynamicRegistration: NotRequired[bool]
    """ Did change configuration notification supports dynamic registration. """


class DidChangeWatchedFilesClientCapabilities(TypedDict):
    dynamicRegistration: NotRequired[bool]
    """ Did change watched files notification supports dynamic registration. Please note
    that the current protocol doesn't support static configuration for file changes
    from the server side. """
    relativePatternSupport: NotRequired[bool]
    """ Whether the client has support for {@link  RelativePattern relative pattern}
    or not.

    @since 3.17.0 """


class WorkspaceSymbolClientCapabilities(TypedDict):
    """Client capabilities for a {@link WorkspaceSymbolRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Symbol request supports dynamic registration. """
    symbolKind: NotRequired["__WorkspaceSymbolClientCapabilities_symbolKind_Type_1"]
    """ Specific capabilities for the `SymbolKind` in the `workspace/symbol` request. """
    tagSupport: NotRequired["__WorkspaceSymbolClientCapabilities_tagSupport_Type_1"]
    """ The client supports tags on `SymbolInformation`.
    Clients supporting tags have to handle unknown tags gracefully.

    @since 3.16.0 """
    resolveSupport: NotRequired["__WorkspaceSymbolClientCapabilities_resolveSupport_Type_1"]
    """ The client support partial workspace symbols. The client will send the
    request `workspaceSymbol/resolve` to the server to resolve additional
    properties.

    @since 3.17.0 """


class ExecuteCommandClientCapabilities(TypedDict):
    """The client capabilities of a {@link ExecuteCommandRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Execute command supports dynamic registration. """


class SemanticTokensWorkspaceClientCapabilities(TypedDict):
    """@since 3.16.0"""

    refreshSupport: NotRequired[bool]
    """ Whether the client implementation supports a refresh request sent from
    the server to the client.

    Note that this event is global and will force the client to refresh all
    semantic tokens currently shown. It should be used with absolute care
    and is useful for situation where a server for example detects a project
    wide change that requires such a calculation. """


class CodeLensWorkspaceClientCapabilities(TypedDict):
    """@since 3.16.0"""

    refreshSupport: NotRequired[bool]
    """ Whether the client implementation supports a refresh request sent from the
    server to the client.

    Note that this event is global and will force the client to refresh all
    code lenses currently shown. It should be used with absolute care and is
    useful for situation where a server for example detect a project wide
    change that requires such a calculation. """


class FileOperationClientCapabilities(TypedDict):
    """Capabilities relating to events from file operations by the user in the client.

    These events do not come from the file system, they come from user operations
    like renaming a file in the UI.

    @since 3.16.0
    """

    dynamicRegistration: NotRequired[bool]
    """ Whether the client supports dynamic registration for file requests/notifications. """
    didCreate: NotRequired[bool]
    """ The client has support for sending didCreateFiles notifications. """
    willCreate: NotRequired[bool]
    """ The client has support for sending willCreateFiles requests. """
    didRename: NotRequired[bool]
    """ The client has support for sending didRenameFiles notifications. """
    willRename: NotRequired[bool]
    """ The client has support for sending willRenameFiles requests. """
    didDelete: NotRequired[bool]
    """ The client has support for sending didDeleteFiles notifications. """
    willDelete: NotRequired[bool]
    """ The client has support for sending willDeleteFiles requests. """


class InlineValueWorkspaceClientCapabilities(TypedDict):
    """Client workspace capabilities specific to inline values.

    @since 3.17.0
    """

    refreshSupport: NotRequired[bool]
    """ Whether the client implementation supports a refresh request sent from the
    server to the client.

    Note that this event is global and will force the client to refresh all
    inline values currently shown. It should be used with absolute care and is
    useful for situation where a server for example detects a project wide
    change that requires such a calculation. """


class InlayHintWorkspaceClientCapabilities(TypedDict):
    """Client workspace capabilities specific to inlay hints.

    @since 3.17.0
    """

    refreshSupport: NotRequired[bool]
    """ Whether the client implementation supports a refresh request sent from
    the server to the client.

    Note that this event is global and will force the client to refresh all
    inlay hints currently shown. It should be used with absolute care and
    is useful for situation where a server for example detects a project wide
    change that requires such a calculation. """


class DiagnosticWorkspaceClientCapabilities(TypedDict):
    """Workspace client capabilities specific to diagnostic pull requests.

    @since 3.17.0
    """

    refreshSupport: NotRequired[bool]
    """ Whether the client implementation supports a refresh request sent from
    the server to the client.

    Note that this event is global and will force the client to refresh all
    pulled diagnostics currently shown. It should be used with absolute care and
    is useful for situation where a server for example detects a project wide
    change that requires such a calculation. """


class TextDocumentSyncClientCapabilities(TypedDict):
    dynamicRegistration: NotRequired[bool]
    """ Whether text document synchronization supports dynamic registration. """
    willSave: NotRequired[bool]
    """ The client supports sending will save notifications. """
    willSaveWaitUntil: NotRequired[bool]
    """ The client supports sending a will save request and
    waits for a response providing text edits which will
    be applied to the document before it is saved. """
    didSave: NotRequired[bool]
    """ The client supports did save notifications. """


class CompletionClientCapabilities(TypedDict):
    """Completion client capabilities"""

    dynamicRegistration: NotRequired[bool]
    """ Whether completion supports dynamic registration. """
    completionItem: NotRequired["__CompletionClientCapabilities_completionItem_Type_1"]
    """ The client supports the following `CompletionItem` specific
    capabilities. """
    completionItemKind: NotRequired["__CompletionClientCapabilities_completionItemKind_Type_1"]
    insertTextMode: NotRequired["InsertTextMode"]
    """ Defines how the client handles whitespace and indentation
    when accepting a completion item that uses multi line
    text in either `insertText` or `textEdit`.

    @since 3.17.0 """
    contextSupport: NotRequired[bool]
    """ The client supports to send additional context information for a
    `textDocument/completion` request. """
    completionList: NotRequired["__CompletionClientCapabilities_completionList_Type_1"]
    """ The client supports the following `CompletionList` specific
    capabilities.

    @since 3.17.0 """


class HoverClientCapabilities(TypedDict):
    dynamicRegistration: NotRequired[bool]
    """ Whether hover supports dynamic registration. """
    contentFormat: NotRequired[list["MarkupKind"]]
    """ Client supports the following content formats for the content
    property. The order describes the preferred format of the client. """


class SignatureHelpClientCapabilities(TypedDict):
    """Client Capabilities for a {@link SignatureHelpRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether signature help supports dynamic registration. """
    signatureInformation: NotRequired["__SignatureHelpClientCapabilities_signatureInformation_Type_1"]
    """ The client supports the following `SignatureInformation`
    specific properties. """
    contextSupport: NotRequired[bool]
    """ The client supports to send additional context information for a
    `textDocument/signatureHelp` request. A client that opts into
    contextSupport will also support the `retriggerCharacters` on
    `SignatureHelpOptions`.

    @since 3.15.0 """


class DeclarationClientCapabilities(TypedDict):
    """@since 3.14.0"""

    dynamicRegistration: NotRequired[bool]
    """ Whether declaration supports dynamic registration. If this is set to `true`
    the client supports the new `DeclarationRegistrationOptions` return value
    for the corresponding server capability as well. """
    linkSupport: NotRequired[bool]
    """ The client supports additional metadata in the form of declaration links. """


class DefinitionClientCapabilities(TypedDict):
    """Client Capabilities for a {@link DefinitionRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether definition supports dynamic registration. """
    linkSupport: NotRequired[bool]
    """ The client supports additional metadata in the form of definition links.

    @since 3.14.0 """


class TypeDefinitionClientCapabilities(TypedDict):
    """Since 3.6.0"""

    dynamicRegistration: NotRequired[bool]
    """ Whether implementation supports dynamic registration. If this is set to `true`
    the client supports the new `TypeDefinitionRegistrationOptions` return value
    for the corresponding server capability as well. """
    linkSupport: NotRequired[bool]
    """ The client supports additional metadata in the form of definition links.

    Since 3.14.0 """


class ImplementationClientCapabilities(TypedDict):
    """@since 3.6.0"""

    dynamicRegistration: NotRequired[bool]
    """ Whether implementation supports dynamic registration. If this is set to `true`
    the client supports the new `ImplementationRegistrationOptions` return value
    for the corresponding server capability as well. """
    linkSupport: NotRequired[bool]
    """ The client supports additional metadata in the form of definition links.

    @since 3.14.0 """


class ReferenceClientCapabilities(TypedDict):
    """Client Capabilities for a {@link ReferencesRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether references supports dynamic registration. """


class DocumentHighlightClientCapabilities(TypedDict):
    """Client Capabilities for a {@link DocumentHighlightRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether document highlight supports dynamic registration. """


class DocumentSymbolClientCapabilities(TypedDict):
    """Client Capabilities for a {@link DocumentSymbolRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether document symbol supports dynamic registration. """
    symbolKind: NotRequired["__DocumentSymbolClientCapabilities_symbolKind_Type_1"]
    """ Specific capabilities for the `SymbolKind` in the
    `textDocument/documentSymbol` request. """
    hierarchicalDocumentSymbolSupport: NotRequired[bool]
    """ The client supports hierarchical document symbols. """
    tagSupport: NotRequired["__DocumentSymbolClientCapabilities_tagSupport_Type_1"]
    """ The client supports tags on `SymbolInformation`. Tags are supported on
    `DocumentSymbol` if `hierarchicalDocumentSymbolSupport` is set to true.
    Clients supporting tags have to handle unknown tags gracefully.

    @since 3.16.0 """
    labelSupport: NotRequired[bool]
    """ The client supports an additional label presented in the UI when
    registering a document symbol provider.

    @since 3.16.0 """


class CodeActionClientCapabilities(TypedDict):
    """The Client Capabilities of a {@link CodeActionRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether code action supports dynamic registration. """
    codeActionLiteralSupport: NotRequired["__CodeActionClientCapabilities_codeActionLiteralSupport_Type_1"]
    """ The client support code action literals of type `CodeAction` as a valid
    response of the `textDocument/codeAction` request. If the property is not
    set the request can only return `Command` literals.

    @since 3.8.0 """
    isPreferredSupport: NotRequired[bool]
    """ Whether code action supports the `isPreferred` property.

    @since 3.15.0 """
    disabledSupport: NotRequired[bool]
    """ Whether code action supports the `disabled` property.

    @since 3.16.0 """
    dataSupport: NotRequired[bool]
    """ Whether code action supports the `data` property which is
    preserved between a `textDocument/codeAction` and a
    `codeAction/resolve` request.

    @since 3.16.0 """
    resolveSupport: NotRequired["__CodeActionClientCapabilities_resolveSupport_Type_1"]
    """ Whether the client supports resolving additional code action
    properties via a separate `codeAction/resolve` request.

    @since 3.16.0 """
    honorsChangeAnnotations: NotRequired[bool]
    """ Whether the client honors the change annotations in
    text edits and resource operations returned via the
    `CodeAction#edit` property by for example presenting
    the workspace edit in the user interface and asking
    for confirmation.

    @since 3.16.0 """


class CodeLensClientCapabilities(TypedDict):
    """The client capabilities  of a {@link CodeLensRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether code lens supports dynamic registration. """


class DocumentLinkClientCapabilities(TypedDict):
    """The client capabilities of a {@link DocumentLinkRequest}."""

    dynamicRegistration: NotRequired[bool]
    """ Whether document link supports dynamic registration. """
    tooltipSupport: NotRequired[bool]
    """ Whether the client supports the `tooltip` property on `DocumentLink`.

    @since 3.15.0 """


class DocumentColorClientCapabilities(TypedDict):
    dynamicRegistration: NotRequired[bool]
    """ Whether implementation supports dynamic registration. If this is set to `true`
    the client supports the new `DocumentColorRegistrationOptions` return value
    for the corresponding server capability as well. """





import logging
import os
import pathlib
import subprocess
from typing import cast

from overrides import override

from solidlsp.ls import SolidLanguageServer
from solidlsp.ls_config import LanguageServerConfig
from solidlsp.lsp_protocol_handler.lsp_types import InitializeParams
from solidlsp.lsp_protocol_handler.server import ProcessLaunchInfo
from solidlsp.settings import SolidLSPSettings

log = logging.getLogger(__name__)


class Gopls(SolidLanguageServer):
    """
    Provides Go specific instantiation of the LanguageServer class using gopls.
    """

    @override
    def is_ignored_dirname(self, dirname: str) -> bool:
        # For Go projects, we should ignore:
        # - vendor: third-party dependencies vendored into the project
        # - node_modules: if the project has JavaScript components
        # - dist/build: common output directories
        return super().is_ignored_dirname(dirname) or dirname in ["vendor", "node_modules", "dist", "build"]

    @staticmethod
    def _determine_log_level(line: str) -> int:
        """Classify gopls stderr output to avoid false-positive errors."""
        line_lower = line.lower()

        # File discovery messages that are not actual errors
        if any(
            [
                "discover.go:" in line_lower,
                "walker.go:" in line_lower,
                "walking of {file://" in line_lower,
                "bus: -> discover" in line_lower,
            ]
        ):
            return logging.DEBUG

        return SolidLanguageServer._determine_log_level(line)

    @staticmethod
    def _get_go_version() -> str | None:
        """Get the installed Go version or None if not found."""
        try:
            result = subprocess.run(["go", "version"], capture_output=True, text=True, check=False)
            if result.returncode == 0:
                return result.stdout.strip()
        except FileNotFoundError:
            return None
        return None

    @staticmethod
    def _get_gopls_version() -> str | None:
        """Get the installed gopls version or None if not found."""
        try:
            result = subprocess.run(["gopls", "version"], capture_output=True, text=True, check=False)
            if result.returncode == 0:
                return result.stdout.strip()
        except FileNotFoundError:
            return None
        return None

    @staticmethod
    def _setup_runtime_dependency() -> bool:
        """
        Check if required Go runtime dependencies are available.
        Raises RuntimeError with helpful message if dependencies are missing.
        """
        go_version = Gopls._get_go_version()
        if not go_version:
            raise RuntimeError(
                "Go is not installed. Please install Go from https://golang.org/doc/install and make sure it is added to your PATH."
            )

        gopls_version = Gopls._get_gopls_version()
        if not gopls_version:
            raise RuntimeError(
                "Found a Go version but gopls is not installed.\n"
                "Please install gopls as described in https://pkg.go.dev/golang.org/x/tools/gopls#section-readme\n\n"
                "After installation, make sure it is added to your PATH (it might be installed in a different location than Go)."
            )

        return True

    def __init__(self, config: LanguageServerConfig, repository_root_path: str, solidlsp_settings: SolidLSPSettings):
        self._setup_runtime_dependency()

        super().__init__(config, repository_root_path, ProcessLaunchInfo(cmd="gopls", cwd=repository_root_path), "go", solidlsp_settings)
        self.request_id = 0

    def _get_initialize_params(self, repository_absolute_path: str) -> InitializeParams:
        """
        Returns the initialize params for the Go Language Server.
        """
        root_uri = pathlib.Path(repository_absolute_path).as_uri()
        initialize_params: dict[str, object] = {
            "locale": "en",
            "capabilities": {
                "textDocument": {
                    "synchronization": {"didSave": True, "dynamicRegistration": True},
                    "definition": {"dynamicRegistration": True},
                    "documentSymbol": {
                        "dynamicRegistration": True,
                        "hierarchicalDocumentSymbolSupport": True,
                        "symbolKind": {"valueSet": list(range(1, 27))},
                    },
                },
                "workspace": {"workspaceFolders": True, "didChangeConfiguration": {"dynamicRegistration": True}},
            },
            "processId": os.getpid(),
            "rootPath": repository_absolute_path,
            "rootUri": root_uri,
            "workspaceFolders": [
                {
                    "uri": root_uri,
                    "name": os.path.basename(repository_absolute_path),
                }
            ],
        }

        # Apply gopls-specific settings via initializationOptions
        # Serena applies gopls settings at initialization time via initializationOptions
        # (Access settings directly to avoid extra INFO logging from CustomLSSettings.get.)
        gopls_settings = self._custom_settings.settings.get("gopls_settings")
        if gopls_settings:
            gopls_settings = self._validate_gopls_settings_dict(gopls_settings)

            # Validate JSON-serializability early: initializationOptions is sent over JSON-RPC.
            import json

            self._canonical_json_or_raise(gopls_settings)

            # Log keys only (and at DEBUG) to avoid leaking sensitive values and to reduce startup noise.
            log.debug("Applying gopls settings via initializationOptions: keys=%s", list(gopls_settings.keys()))
            initialize_params["initializationOptions"] = gopls_settings

        return cast(InitializeParams, initialize_params)

    def _validate_gopls_settings_dict(self, gopls_settings: object) -> dict[str, object]:
        if not isinstance(gopls_settings, dict):
            raise TypeError(
                f"gopls_settings must be a dict, got {type(gopls_settings).__name__}. "
                "Expected structure: {'buildFlags': ['-tags=foo'], 'env': {...}, ...}"
            )

        return gopls_settings

    def _canonical_json_or_raise(self, data: object) -> str:
        try:
            import json

            return json.dumps(data, sort_keys=True, separators=(",", ":"))
        except (TypeError, ValueError) as exc:
            raise TypeError(
                "gopls_settings must be JSON-serializable (json.dumps). Use JSON-compatible values (dict/list/str/int/float/bool/null) and prefer string keys."
            ) from exc

    # Environment variables that influence Go build context and affect cached symbols.
    _CACHE_CONTEXT_ENV_KEYS = ("GOFLAGS", "GOOS", "GOARCH", "CGO_ENABLED")

    @override
    def _cache_context_fingerprint(self) -> str | None:
        """
        Compute a deterministic fingerprint of the Go build context.

        The fingerprint includes gopls_settings and selected env vars that affect symbol discovery.
        """
        import hashlib
        import json

        gopls_settings_raw = self._custom_settings.settings.get("gopls_settings")

        gopls_settings: dict | None
        if gopls_settings_raw is None:
            gopls_settings = None
        else:
            # Treat an explicitly empty dict the same as not providing settings at all.
            gopls_settings = self._validate_gopls_settings_dict(gopls_settings_raw) or None

        # Only include env vars that are set to a non-empty value.
        env_subset: dict[str, str] = {}
        for key in self._CACHE_CONTEXT_ENV_KEYS:
            value = os.environ.get(key)
            if value:
                env_subset[key] = value

        # Return None only when BOTH settings and env subset are effectively empty.
        if gopls_settings is None and not env_subset:
            return None

        fingerprint_data: dict[str, object] = {"env": env_subset}
        if gopls_settings is not None:
            fingerprint_data["gopls_settings"] = gopls_settings

        canonical_json = self._canonical_json_or_raise(json, fingerprint_data)

        return hashlib.sha256(canonical_json.encode("utf-8")).hexdigest()[:16]

    def _start_server(self) -> None:
        """Start gopls server process"""

        def register_capability_handler(params: dict) -> None:
            return

        def window_log_message(msg: dict) -> None:
            log.info(f"LSP: window/logMessage: {msg}")

        def do_nothing(params: dict) -> None:
            return

        self.server.on_request("client/registerCapability", register_capability_handler)
        self.server.on_notification("window/logMessage", window_log_message)
        self.server.on_notification("$/progress", do_nothing)
        self.server.on_notification("textDocument/publishDiagnostics", do_nothing)

        log.info("Starting gopls server process")
        self.server.start()
        initialize_params = self._get_initialize_params(self.repository_root_path)

        log.info("Sending initialize request from LSP client to LSP server and awaiting response")
        init_response = self.server.send.initialize(initialize_params)

        # Verify server capabilities
        assert "textDocumentSync" in init_response["capabilities"]
        assert "completionProvider" in init_response["capabilities"]
        assert "definitionProvider" in init_response["capabilities"]

        self.server.notify.initialized({})

        # gopls server is typically ready immediately after initialization
        # (no need to wait for events)



"""LSP file URI 파싱 유틸리티를 제공한다."""

from __future__ import annotations

from pathlib import Path, PureWindowsPath
from urllib.parse import unquote, urlparse

from sari.core.exceptions import ErrorContext, ValidationError


def file_uri_to_repo_relative(uri: str, repo_root: str) -> str:
    """file URI를 repo 상대 경로로 변환한다."""
    parsed = urlparse(uri)
    if parsed.scheme != "file":
        raise ValidationError(ErrorContext(code="ERR_URI_PATH_INVALID", message="file URI만 지원합니다"))

    decoded_path = unquote(parsed.path)
    if _is_windows_uri(parsed.path, parsed.netloc, repo_root):
        return _file_uri_to_repo_relative_windows(parsed=parsed, decoded_path=decoded_path, repo_root=repo_root)

    absolute_path = Path(decoded_path).resolve()
    repo_path = Path(repo_root).resolve()
    try:
        return str(absolute_path.relative_to(repo_path).as_posix())
    except ValueError as exc:
        raise ValidationError(ErrorContext(code="ERR_URI_PATH_INVALID", message="uri가 repo 범위를 벗어났습니다")) from exc


def _is_windows_uri(uri_path: str, uri_netloc: str, repo_root: str) -> bool:
    """Windows 스타일 URI/경로 여부를 판별한다."""
    if uri_netloc.strip() != "":
        return True
    if len(uri_path) >= 4 and uri_path[0] == "/" and uri_path[2] == ":":
        return True
    if len(uri_path) >= 3 and uri_path[1] == ":":
        return True
    if ":\\" in repo_root or ":/" in repo_root:
        return True
    return False


def _file_uri_to_repo_relative_windows(parsed, decoded_path: str, repo_root: str) -> str:  # type: ignore[no-untyped-def]
    """Windows file URI를 repo 상대 경로로 변환한다."""
    normalized = decoded_path.replace("/", "\\")
    # file:///C:/path -> C:\path
    if len(normalized) >= 4 and normalized[0] == "\\" and normalized[2] == ":":
        normalized = normalized[1:]
    # file://server/share/path -> \\server\share\path
    if parsed.netloc.strip() != "":
        normalized = f"\\\\{parsed.netloc}{normalized}"

    absolute_path = PureWindowsPath(normalized)
    repo_path = PureWindowsPath(repo_root)
    absolute_parts = tuple(part.lower() for part in absolute_path.parts)
    repo_parts = tuple(part.lower() for part in repo_path.parts)
    if len(repo_parts) == 0 or absolute_parts[: len(repo_parts)] != repo_parts:
        raise ValidationError(ErrorContext(code="ERR_URI_PATH_INVALID", message="uri가 repo 범위를 벗어났습니다"))
    relative_parts = absolute_path.parts[len(repo_path.parts) :]
    return PureWindowsPath(*relative_parts).as_posix()



"""LSP 기반 심볼 해석 레이어를 구현한다."""

from __future__ import annotations

import logging
from typing import Protocol

from sari.core.exceptions import DaemonError, ErrorContext, ValidationError
from sari.core.models import CandidateFileDTO, SearchErrorDTO, SearchItemDTO
from sari.db.repositories.symbol_cache_repository import SymbolCacheRepository
from sari.lsp.hub import LspHub
from sari.lsp.path_normalizer import normalize_location_to_repo_relative, normalize_repo_relative_path
from sari.search.error_policy import classify_search_error
from solidlsp.ls_config import Language
from solidlsp.ls_exceptions import SolidLSPException

log = logging.getLogger(__name__)


class LspQueryBackend(Protocol):
    """심볼 조회 백엔드 프로토콜을 정의한다."""

    def query_document_symbols(
        self,
        repo_root: str,
        language: Language,
        relative_paths: list[str],
        query: str,
        limit: int,
    ) -> list[SearchItemDTO]:
        """파일 단위 documentSymbol 조회 결과를 반환한다."""


class SolidLspQueryBackend:
    """solidlsp를 이용한 실제 심볼 조회 백엔드다."""

    def __init__(self, hub: LspHub) -> None:
        """LSP Hub 의존성을 주입한다."""
        self._hub = hub

    def query_document_symbols(
        self,
        repo_root: str,
        language: Language,
        relative_paths: list[str],
        query: str,
        limit: int,
    ) -> list[SearchItemDTO]:
        """documentSymbol 결과를 SearchItem DTO로 변환한다."""
        try:
            self._hub.ensure_healthy(language=language, repo_root=repo_root)  # type: ignore[attr-defined]
        except DaemonError as exc:
            if exc.context.code != "ERR_LSP_UNHEALTHY":
                raise
            self._hub.restart_if_unhealthy(language=language, repo_root=repo_root)  # type: ignore[attr-defined]
        lsp = self._hub.get_or_start(language=language, repo_root=repo_root)
        normalized_query = query.strip().lower()
        items: list[SearchItemDTO] = []
        for relative_path in relative_paths:
            normalized_relative_path = normalize_repo_relative_path(relative_path)
            try:
                raw_symbols = list(lsp.request_document_symbols(normalized_relative_path).iter_symbols())
            except SolidLSPException as exc:
                message = str(exc)
                if "ERR_LSP_SYNC_OPEN_FAILED" in message:
                    raise DaemonError(
                        ErrorContext(
                            code="ERR_LSP_SYNC_OPEN_FAILED",
                            message=f"LSP 문서 open 동기화 실패(path={normalized_relative_path}): {message}",
                        )
                    ) from exc
                if "ERR_LSP_SYNC_CHANGE_FAILED" in message:
                    raise DaemonError(
                        ErrorContext(
                            code="ERR_LSP_SYNC_CHANGE_FAILED",
                            message=f"LSP 문서 change 동기화 실패(path={normalized_relative_path}): {message}",
                        )
                    ) from exc
                raise DaemonError(
                    ErrorContext(
                        code="ERR_LSP_DOCUMENT_SYMBOL_FAILED",
                        message=f"documentSymbol 요청 실패(path={normalized_relative_path}): {message}",
                    )
                ) from exc
            for symbol in raw_symbols:
                if not isinstance(symbol, dict):
                    continue
                symbol_name = symbol.get("name")
                if not isinstance(symbol_name, str):
                    continue
                if normalized_query != "" and normalized_query not in symbol_name.lower():
                    continue
                resolved_relative_path = normalized_relative_path
                location = symbol.get("location")
                if isinstance(location, dict):
                    try:
                        resolved_relative_path = normalize_location_to_repo_relative(
                            location=location,
                            fallback_relative_path=normalized_relative_path,
                            repo_root=repo_root,
                        )
                    except ValidationError:
                        resolved_relative_path = normalized_relative_path
                symbol_kind = symbol.get("kind")
                symbol_key = symbol.get("symbol_key")
                parent_symbol_key = symbol.get("parent_symbol_key")
                depth = symbol.get("depth")
                container_name = symbol.get("container_name")
                items.append(
                    SearchItemDTO(
                        item_type="symbol",
                        repo=repo_root,
                        relative_path=str(resolved_relative_path),
                        score=1.0,
                        source="lsp",
                        name=symbol_name,
                        kind=str(symbol_kind) if symbol_kind is not None else None,
                        symbol_key=str(symbol_key) if isinstance(symbol_key, str) and symbol_key.strip() != "" else None,
                        parent_symbol_key=(
                            str(parent_symbol_key)
                            if isinstance(parent_symbol_key, str) and parent_symbol_key.strip() != ""
                            else None
                        ),
                        depth=int(depth) if isinstance(depth, int) else 0,
                        container_name=(
                            str(container_name) if isinstance(container_name, str) and container_name.strip() != "" else None
                        ),
                    )
                )
                if len(items) >= limit:
                    return items[:limit]
        return items


class SymbolResolveService:
    """후보 파일을 LSP 심볼로 해석하는 서비스다."""

    def __init__(
        self,
        hub: LspHub,
        cache_repo: SymbolCacheRepository,
        backend: LspQueryBackend | None = None,
    ) -> None:
        """LSP Hub/캐시 저장소/백엔드를 주입한다."""
        self._hub = hub
        self._cache_repo = cache_repo
        self._backend = backend or SolidLspQueryBackend(hub)

    def resolve(self, candidates: list[CandidateFileDTO], query: str, limit: int) -> tuple[list[SearchItemDTO], list[SearchErrorDTO]]:
        """후보 파일 기반으로 LSP 심볼 결과를 생성한다."""
        items: list[SearchItemDTO] = []
        errors: list[SearchErrorDTO] = []

        miss_candidates: list[CandidateFileDTO] = []
        grouped_languages: dict[tuple[str, Language], list[str]] = {}

        for candidate in candidates:
            cached_items = self._cache_repo.get_cached_items(
                repo_root=candidate.repo_root,
                relative_path=candidate.relative_path,
                query=query,
                file_hash=candidate.file_hash,
            )
            if cached_items is not None:
                for cached in cached_items:
                    items.append(cached)
                    if len(items) >= limit:
                        return items[:limit], errors
                continue

            miss_candidates.append(candidate)
            try:
                language = self._hub.resolve_language(candidate.relative_path)
            except DaemonError as exc:
                log.warning(
                    "LSP 언어 매핑 실패(repo=%s,path=%s,query=%s): %s",
                    candidate.repo_root,
                    candidate.relative_path,
                    query,
                    exc.context.message,
                )
                errors.append(
                    SearchErrorDTO(
                        code=exc.context.code,
                        message=exc.context.message,
                        severity=classify_search_error(exc.context.code),
                        origin="symbol_resolve",
                    )
                )
                continue
            grouped_languages.setdefault((candidate.repo_root, language), []).append(candidate.relative_path)

        if len(miss_candidates) == 0:
            return items[:limit], errors

        miss_path_set = {(candidate.repo_root, candidate.relative_path) for candidate in miss_candidates}
        miss_map = {(candidate.repo_root, candidate.relative_path): candidate for candidate in miss_candidates}

        per_path_items: dict[tuple[str, str], list[SearchItemDTO]] = {}
        for grouped_key, relative_paths in grouped_languages.items():
            repo_root, language = grouped_key
            try:
                resolved = self._query_symbols(
                    repo_root=repo_root,
                    language=language,
                    relative_paths=relative_paths,
                    query=query,
                    limit=limit,
                )
            except DaemonError as exc:
                log.warning("LSP 심볼 조회 도메인 실패(repo=%s,language=%s,query=%s): %s", repo_root, language.value, query, exc.context.message)
                errors.append(
                    SearchErrorDTO(
                        code=exc.context.code,
                        message=exc.context.message,
                        severity=classify_search_error(exc.context.code),
                        origin="symbol_resolve",
                    )
                )
                continue
            except ValidationError as exc:
                log.warning("LSP URI 파싱 실패(repo=%s,language=%s,query=%s): %s", repo_root, language.value, query, exc.context.message)
                errors.append(
                    SearchErrorDTO(
                        code=exc.context.code,
                        message=exc.context.message,
                        severity=classify_search_error(exc.context.code),
                        origin="symbol_resolve",
                    )
                )
                continue
            except (RuntimeError, OSError, ValueError, TypeError) as exc:
                # 운영에서 원인 추적이 가능하도록 예외 메시지를 명시적으로 포함한다.
                log.exception("LSP 심볼 조회 실패(repo=%s, language=%s, query=%s): %s", repo_root, language.value, query, exc)
                code = "ERR_LSP_QUERY_FAILED"
                errors.append(
                    SearchErrorDTO(
                        code=code,
                        message=f"LSP 심볼 조회 중 오류가 발생했습니다: {exc}",
                        severity=classify_search_error(code),
                        origin="symbol_resolve",
                    )
                )
                continue

            for item in resolved:
                path_key = (item.repo, item.relative_path)
                if path_key not in miss_path_set:
                    continue
                bucket = per_path_items.setdefault(path_key, [])
                bucket.append(item)

        for path_key, candidate in miss_map.items():
            path_items = per_path_items.get(path_key, [])
            materialized_items = [
                SearchItemDTO(
                    item_type=item.item_type,
                    repo=item.repo,
                    relative_path=item.relative_path,
                    score=item.score,
                    source=item.source,
                    name=item.name,
                    kind=item.kind,
                    content_hash=candidate.file_hash,
                    rrf_score=item.rrf_score,
                    importance_score=item.importance_score,
                    hierarchy_score=item.hierarchy_score,
                    hierarchy_norm_score=item.hierarchy_norm_score,
                    symbol_key=item.symbol_key,
                    parent_symbol_key=item.parent_symbol_key,
                    depth=item.depth,
                    container_name=item.container_name,
                    ranking_components=item.ranking_components,
                    vector_score=item.vector_score,
                    final_score=item.final_score,
                )
                for item in path_items
            ]
            self._cache_repo.upsert_items(
                repo_root=candidate.repo_root,
                relative_path=candidate.relative_path,
                query=query,
                file_hash=candidate.file_hash,
                items=materialized_items,
            )
            for item in materialized_items:
                items.append(item)
                if len(items) >= limit:
                    return items[:limit], errors

        return items[:limit], errors

    def _query_symbols(
        self,
        repo_root: str,
        language: Language,
        relative_paths: list[str],
        query: str,
        limit: int,
    ) -> list[SearchItemDTO]:
        """백엔드 호환 모드(document/workspace)를 모두 지원해 심볼을 조회한다."""
        if hasattr(self._backend, "query_document_symbols"):
            return self._backend.query_document_symbols(
                repo_root=repo_root,
                language=language,
                relative_paths=relative_paths,
                query=query,
                limit=limit,
            )
        # 이전 테스트 더블 호환을 위한 workspace 쿼리 fallback이다.
        if hasattr(self._backend, "query_workspace_symbol"):
            queried = self._backend.query_workspace_symbol(  # type: ignore[attr-defined]
                repo_root=repo_root,
                language=language,
                query=query,
            )
            return queried[:limit]
        raise ValueError("symbol backend contract is invalid")



"""MCP 파일 수집 도구(scan_once/list_files/read_file/index_file)를 제공한다."""

from __future__ import annotations

from sari.core.exceptions import CollectionError
from sari.core.models import ErrorResponseDTO
from sari.db.repositories.workspace_repository import WorkspaceRepository
from sari.mcp.tools.arg_parser import parse_non_empty_string, parse_non_negative_int, parse_optional_int, parse_optional_string, parse_positive_int
from sari.mcp.tools.admin_tools import validate_repo_argument
from sari.mcp.tools.pack1 import Pack1MetaDTO, pack1_error, pack1_success
from sari.services.collection.ports import CollectionScanPort


def _collection_error_response(exc: CollectionError) -> dict[str, object]:
    """CollectionError를 pack1 명시적 오류 응답으로 변환한다."""
    return pack1_error(ErrorResponseDTO(code=exc.context.code, message=exc.context.message))


class ScanOnceTool:
    """scan_once MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, collection_service: CollectionScanPort) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._collection_service = collection_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """저장소 전체 1회 스캔을 실행한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        repo = str(arguments["repo"])

        try:
            result = self._collection_service.scan_once(repo_root=repo)
        except CollectionError as exc:
            return _collection_error_response(exc)

        item = {
            "repo": repo,
            "scanned_count": result.scanned_count,
            "indexed_count": result.indexed_count,
            "deleted_count": result.deleted_count,
        }
        return pack1_success(
            {
                "items": [item],
                "meta": Pack1MetaDTO(
                    candidate_count=result.scanned_count,
                    resolved_count=result.indexed_count,
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )


class ListFilesTool:
    """list_files MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, collection_service: CollectionScanPort) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._collection_service = collection_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """파일 목록을 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        repo = str(arguments["repo"])

        limit_raw, limit_error = parse_positive_int(arguments=arguments, key="limit", default=20)
        if limit_error is not None:
            return pack1_error(limit_error)
        prefix = parse_optional_string(arguments=arguments, key="prefix")

        try:
            items = self._collection_service.list_files(repo_root=repo, limit=limit_raw, prefix=prefix)
        except CollectionError as exc:
            return _collection_error_response(exc)

        return pack1_success(
            {
                "items": items,
                "meta": Pack1MetaDTO(
                    candidate_count=len(items),
                    resolved_count=len(items),
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )


class ReadFileTool:
    """read_file MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, collection_service: CollectionScanPort) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._collection_service = collection_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """파일 내용을 pack1 형식으로 반환한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        repo = str(arguments["repo"])

        relative_path_raw, path_error = parse_non_empty_string(arguments=arguments, key="relative_path")
        if path_error is not None:
            return pack1_error(path_error)
        offset_raw, offset_error = parse_non_negative_int(arguments=arguments, key="offset", default=0)
        if offset_error is not None:
            return pack1_error(offset_error)
        limit_raw, limit_error = parse_optional_int(arguments=arguments, key="limit", default=300)
        if limit_error is not None:
            return pack1_error(limit_error)

        try:
            result = self._collection_service.read_file(
                repo_root=repo,
                relative_path=relative_path_raw,
                offset=offset_raw,
                limit=limit_raw,
            )
        except CollectionError as exc:
            return _collection_error_response(exc)

        item = {
            "relative_path": result.relative_path,
            "content": result.content,
            "start_line": result.start_line,
            "end_line": result.end_line,
            "source": result.source,
            "total_lines": result.total_lines,
            "is_truncated": result.is_truncated,
            "next_offset": result.next_offset,
        }
        return pack1_success(
            {
                "items": [item],
                "meta": Pack1MetaDTO(
                    candidate_count=1,
                    resolved_count=1,
                    cache_hit=result.source == "l2",
                    errors=[],
                ).to_dict(),
            }
        )


class IndexFileTool:
    """index_file MCP 도구를 처리한다."""

    def __init__(self, workspace_repo: WorkspaceRepository, collection_service: CollectionScanPort) -> None:
        """필요 의존성을 주입한다."""
        self._workspace_repo = workspace_repo
        self._collection_service = collection_service

    def call(self, arguments: dict[str, object]) -> dict[str, object]:
        """단일 파일 증분 인덱싱을 실행한다."""
        error = validate_repo_argument(arguments=arguments, workspace_repo=self._workspace_repo)
        if error is not None:
            return pack1_error(error)
        repo = str(arguments["repo"])

        relative_path_raw, path_error = parse_non_empty_string(arguments=arguments, key="relative_path")
        if path_error is not None:
            return pack1_error(path_error)

        try:
            result = self._collection_service.index_file(repo_root=repo, relative_path=relative_path_raw)
        except CollectionError as exc:
            return _collection_error_response(exc)

        item = {
            "repo": repo,
            "relative_path": relative_path_raw,
            "scanned_count": result.scanned_count,
            "indexed_count": result.indexed_count,
            "deleted_count": result.deleted_count,
        }
        return pack1_success(
            {
                "items": [item],
                "meta": Pack1MetaDTO(
                    candidate_count=1,
                    resolved_count=result.indexed_count,
                    cache_hit=None,
                    errors=[],
                ).to_dict(),
            }
        )
